'''
This module contains gui helper functions
'''

import matplotlib.pyplot as plt
from matplotlib.widgets import RadioButtons
import matplotlib.cm as cm
from matplotlib.colorbar import Colorbar
import matplotlib.colors as mpl_col
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg
from matplotlib.figure import Figure
import matplotlib.gridspec as gridspec
from matplotlib.ticker import NullFormatter, FormatStrFormatter, LogLocator, \
                              SymmetricalLogLocator, FuncFormatter # noqa

import numpy as np
import numpy.typing as npt
import sys

from qtpy import QtWidgets, QtCore
import pyqtgraph as pg

from . import utils

# Default values for interactive parameter widgets of relaxation.FitWindow
widget_defaults = {
    'Orbach': {
        'u_eff': {
            'min': 0.,
            'max': 3000.,
            'valinit': 1500.,
            'step': 1,
            'decimals': 2
        },
        'A': {
            'min': -30.,
            'max': 30.,
            'valinit': -11.,
            'step': 0.01,
            'decimals': 3
        }
    },
    'Raman': {
        'R': {
            'min': -30.,
            'max': 30.,
            'valinit': -6.,
            'step': 0.01,
            'decimals': 3
        },
        'n': {
            'min': 0,
            'max': 20.,
            'valinit': 3.,
            'step': 0.01,
            'decimals': 3
        }
    },
    'QTM': {
        'Q': {
            'min': -6.,
            'max': 6.,
            'valinit': 1.,
            'step': 0.01,
            'decimals': 4
        }
    },
    'Direct': {
        'D': {
            'min': -6.,
            'max': 6.,
            'valinit': 0.1,
            'step': 0.01,
            'decimals': 4
        }
    },
    'FD-QTM': {
        'Q': {
            'min': -30.,
            'max': 30.,
            'valinit': 0.1,
            'step': 0.01,
            'decimals': 4
        },
        'Q_H': {
            'min': -30.,
            'max': 30.,
            'valinit': 0.1,
            'step': 0.01,
            'decimals': 4
        },
        'p': {
            'min': -30.,
            'max': 30.,
            'valinit': 0.1,
            'step': 0.01,
            'decimals': 4
        }
    },
    'Raman-II': {
        'C': {
            'min': -30.,
            'max': 30.,
            'valinit': 0.1,
            'step': 0.01,
            'decimals': 4
        },
        'm': {
            'min': -30.,
            'max': 30.,
            'valinit': 0.1,
            'step': 0.01,
            'decimals': 4
        }
    },
    'Constant': {
        'Ct': {
            'min': -30.,
            'max': 10.,
            'valinit': 0.1,
            'step': 0.01,
            'decimals': 4
        }
    },
    'Brons-Van-Vleck * Raman-II': {
        'e': {
            'min': -30.,
            'max': 30.,
            'valinit': 0.1,
            'step': 0.01,
            'decimals': 4
        },
        'f': {
            'min': -30.,
            'max': 30.,
            'valinit': 0.1,
            'step': 0.01,
            'decimals': 4
        },
        'C': {
            'min': -30.,
            'max': 30.,
            'valinit': 0.1,
            'step': 0.01,
            'decimals': 4
        },
        'm': {
            'min': -30.,
            'max': 30.,
            'valinit': 0.1,
            'step': 0.01,
            'decimals': 4
        },
    },
    'Brons-Van-Vleck * Constant': {
        'e': {
            'min': -30.,
            'max': 30.,
            'valinit': 0.1,
            'step': 0.01,
            'decimals': 4
        },
        'f': {
            'min': -30.,
            'max': 30.,
            'valinit': 0.1,
            'step': 0.01,
            'decimals': 4
        },
        'Ct': {
            'min': -30.,
            'max': 30.,
            'valinit': 0.1,
            'step': 0.01,
            'decimals': 4
        }
    }
}


def min_max_ticks_with_zero(values: list[float],
                            nticks: int) -> tuple[list[float], float]:
    '''
    Calculates tick positions including zero given a specified number of
    ticks either size of zero

    Parameters
    ----------
    values: list[float]
        Values plotted on this axis e.g. y-values or x-values
    n_ticks: int
        Number of ticks to produce either side of zero.
        i.e. total number of ticks is 2*n_ticks + 1

    Returns
    -------
    list[float]
        Tick positions
    float
        Maximum tick value
    '''

    # Extra tick for zero
    nticks += 1

    lowticks = np.linspace(-np.max(np.abs(values)), 0, nticks)
    highticks = np.linspace(np.max(np.abs(values)), 0, nticks)
    ticks = np.append(np.append(lowticks[:-1], [0.0]), np.flip(highticks[:-1]))

    return ticks, np.max(np.abs(values))


def calc_y_rate_lims(rates: list[float],
                     rate_err: list[float] = []) -> tuple[float, float]:
    '''
    Defines rate plot y limits as 10^integer

    Parameters
    ----------
    rates: list[float]
        Relaxation rates in s^-1
    rate_err: list[float], default []
        Error on rate, upper then lower\n
        Shape (n_rates, 2)

    Returns
    -------
    float
        Upper tick position
    float
        Lower tick position
    '''
    # Define limits of y axis
    # Upper limit from rounding up to nearest power of 10
    # Lower from rounding down to nearest power of 10

    if isinstance(rate_err, list):
        rate_err = np.asarray(rate_err)

    if not len(rate_err):
        rate_err = np.zeros([2, len(rates)])
    y_lower = 10**np.floor(
        np.log10(
            np.nanmin(
                [rates, rates + rate_err[1, :], rates - rate_err[0, :]]
            )
        )
    )
    y_upper = 10**np.ceil(
        np.log10(
            np.nanmax(
                [rates, rates + rate_err[1, :], rates - rate_err[0, :]]
            )
        )
    )

    if np.isnan(y_lower):
        y_lower = y_upper / 10
    if np.isnan(y_upper):
        y_upper = y_lower / 10

    return y_lower, y_upper


def calc_linthresh(x_vals: npt.ArrayLike) -> float:
    '''
    Calculates linthresh for symlog scale using field values.
    Valid only for rate/time versus field plots

    Parameters
    ----------
    x_vals: array_like
        Field values in Oe

    Returns
    -------
    float
        linthresh for symlog scale
    '''

    x_vals = np.asarray(x_vals)

    # Using first value greater than machine eps calculate new threshold
    it = np.argmin(x_vals[np.where(x_vals > np.finfo(float).eps)])
    linthresh = 10**np.floor(
        np.log10(x_vals[np.where(x_vals > np.finfo(float).eps)][it])
    )

    return linthresh


def calc_linscale(x_vals: npt.ArrayLike) -> float:
    '''
    Calculates how much space the linear region takes up on the symlog axis
    Defined here as reciprocal of number of decades spanned by data + 1
    where the +1 accounts for the zero point itself, considered as its own
    decade

    Parameters
    ----------
    x_vals: array_like[float]
        X values

    Returns
    -------
    float
        Width of linear region
    '''

    nz_x_vals = np.log10(x_vals[np.nonzero(x_vals)])
    decs = [np.floor(val) for val in nz_x_vals]

    n_dec = np.max(decs) - np.min(decs) + 1

    return 1 / n_dec


def format_rate_x_y_axes(ax: plt.Axes, rates: list[float],
                         x_vals: list[float],
                         rate_err: list[float] = [],
                         x_type: str = 'temperature') -> None:
    '''
    Wrapper for calc_y_rate_lims and set_rate_xtick_formatting
    Formats both axes of a rate vs T or H plot

    Parameters
    ----------
    ax: plt.Axes
        Axis to modify
    rates: list[float]
        Relaxation rates in s^-1
    x_vals: list[float]
        x values of plot (T or H)
    rate_err: list[float], default []
        Error on rates, upper then lower one per rate.\n
        Shape (n_rates, 2)
    x_type: str, {'field', 'temperature'}
        Type of data, temperature (T) or field (H)

    Returns
    -------
    None
    '''

    y_lower, y_upper = calc_y_rate_lims(rates, rate_err)

    ax.set_ylim([y_lower, y_upper])

    set_rate_xtick_formatting(ax, x_vals, x_type=x_type)

    ax.yaxis.set_minor_locator(LogLocator(base=10, subs='auto'))

    return


def set_rate_xtick_formatting(ax: plt.Axes, x_vals: list[float],
                              x_type: str = 'temperature') -> None:
    '''
    Sets x-tick formatting for rate plot. Enables minor tick labels if <1.1
    order of magnitude spanned by ticks

    Parameters
    ----------
    ax: plt.Axes
        Axis to modify
    x_vals: list[float]
        x values of plot (T or H)
    x_type: str, {'field', 'temperature'}
        Type of data, temperature (T) or field (H)

    Returns
    -------
    None
    '''

    if x_type == 'field':

        # Major ticks
        # Let matplotlib decide decimal values, but stop it
        # converting them to 10^val notation
        ax.xaxis.set_major_formatter(
            FuncFormatter(lambda y, _: '{:g}'.format(y))
        )
        # Disable minor tick labels
        ax.xaxis.set_minor_formatter(NullFormatter())
        # and set minor tick locations
        ax.xaxis.set_minor_locator(
            SymmetricalLogLocator(
                base=10,
                subs=[1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
                linthresh=calc_linthresh(x_vals)
            )
        )
    elif x_type == 'temperature':
        x_vals = np.log10(x_vals)
        # Major ticks
        ax.xaxis.set_major_formatter(FormatStrFormatter('%.0f'))
        # Minor ticks
        if np.max(x_vals) - np.min(x_vals) > 1.05:
            # No minor ticks if > 1 and a bit decades spanned by temperature
            ax.xaxis.set_minor_formatter(NullFormatter())
        else:
            # Add minor tick labels when range is small
            # i.e. only just crosses a decade
            # and make tick lengths equal to major ticks
            ax.xaxis.set_minor_formatter(FormatStrFormatter('%.0f'))
            ax.tick_params('x', length=3.5, width=1, which='major')
            ax.tick_params('x', length=3.5, width=1, which='minor')

    return


def convert_log_ticks_to_lin(ax: pg.graphicsItems.AxisItem.AxisItem,
                             logx_vals: list[float],
                             shift: float = 0.) -> None:
    '''
    Converts logarithmic tick values to linear and adds to pyqtgraph axis

    Shift applies a shift to the ticks, and is neccessary when the data
    has been shifted to accommodate a x_value of 0 (e.g. Field = 0)
    '''

    # Determine size of this item in pixels
    bounds = ax.mapRectFromParent(ax.geometry())
    span = (bounds.topLeft(), bounds.topRight())
    points = list(map(ax.mapToDevice, span))

    lengthInPixels = pg.Point(points[1] - points[0]).length()
    if lengthInPixels == 0:
        return

    # Determine major / minor / subminor axis ticks
    tick_tuple = ax.tickValues(ax.range[0], ax.range[1], lengthInPixels)

    minor_ticks = []
    intermediate_ticks = []

    if len(tick_tuple) >= 1:
        major_tick_vals = tick_tuple[0][1]
        major_ticks = [
            (level + shift, '{:.0f}'.format(10**level))
            for level in major_tick_vals
        ]

    if len(tick_tuple) >= 2:
        minor_tick_vals = tick_tuple[-1][1]

        if np.max(logx_vals) - np.min(logx_vals) < 1.1:
            minor_ticks = [
                (level + shift, '{:.0f}'.format(10**level))
                for level in minor_tick_vals
            ]
        else:
            minor_ticks = [
                (level + shift, '')
                for level in minor_tick_vals
            ]

    if len(tick_tuple) >= 3:
        inter_tick_vals = tick_tuple[1][1]
        intermediate_ticks = [
            (level + shift, '{:.0f}'.format(10**level))
            for level in inter_tick_vals
        ]

    if any(val == 0 for val in logx_vals):
        major_ticks.append((0., '0'))

    ax.setTicks([major_ticks, intermediate_ticks, minor_ticks])

    return


def error_box(message: str, title: str = 'Error') -> None:
    '''
    Creates simple qt error box with specified text and title

    Parameters
    ----------
    message: str
        Text shown in body of window
    title: str, default 'error'
        Text shown in title of window

    Returns
    -------
    None
    '''
    _ = QtWidgets.QApplication(sys.argv)
    msg = QtWidgets.QMessageBox()
    msg.setIcon(QtWidgets.QMessageBox.Critical)
    msg.setText(message)
    msg.setWindowTitle(title)
    msg.exec_()
    sys.exit()


class SampleInfoEntry(QtWidgets.QMainWindow):
    '''
    QT Widget for sample information input (mass and molecular weight)
    '''

    def __init__(self, user_cfg: utils.UserConfig, *args, **kwargs):

        # If user closes, then do not continue as values
        # haven't been entered
        self.close_continue = False

        super(SampleInfoEntry, self).__init__(*args, **kwargs)

        self.setWindowTitle('Sample information')

        self.setAttribute(QtCore.Qt.WA_QuitOnClose)
        self.setAttribute(QtCore.Qt.WA_DeleteOnClose)
        self.setStyleSheet(
            '''
            QMainWindow {
                background-color: white
            }
            '''
        )

        # Minimum Window Size
        self.setMinimumSize(QtCore.QSize(310, 150))

        # Center the window on the screen
        qtRectangle = self.frameGeometry()
        centre_point = QtWidgets.QDesktopWidget().availableGeometry().center()
        qtRectangle.moveCenter(centre_point)
        self.move(qtRectangle.topLeft())
        qtRectangle = self.frameGeometry()
        centerPoint = QtWidgets.QDesktopWidget().availableGeometry().center()
        qtRectangle.moveCenter(centerPoint)
        self.move(qtRectangle.topLeft())

        # Widget to wrap everything
        self.widget = QtWidgets.QWidget(parent=self)
        self.setCentralWidget(self.widget)

        layout = QtWidgets.QVBoxLayout(self.widget)

        # First column, text entry and label for MW
        col_1_widget = QtWidgets.QWidget(parent=self.widget)
        col_1_layout = QtWidgets.QHBoxLayout(col_1_widget)

        self.mw_entry = QtWidgets.QDoubleSpinBox(parent=col_1_widget)
        self.mw_entry.setMaximum(200000)
        self.mw_entry.setDecimals(5)
        self.mw_entry.setButtonSymbols(QtWidgets.QDoubleSpinBox.NoButtons)
        self.mw_entry.setKeyboardTracking(False)

        mw_label = QtWidgets.QLabel(r'Molecular Weight (g mol<sup>-1<\sup>)')
        col_1_layout.addWidget(mw_label)
        col_1_layout.addWidget(self.mw_entry)

        # 2nd column, text entry and label for MW
        col_2_widget = QtWidgets.QWidget(parent=self.widget)
        col_2_layout = QtWidgets.QHBoxLayout(col_2_widget)

        self.mass_entry = QtWidgets.QDoubleSpinBox(parent=col_2_widget)
        self.mass_entry.setMaximum(200000)
        self.mass_entry.setDecimals(5)
        self.mass_entry.setButtonSymbols(QtWidgets.QDoubleSpinBox.NoButtons)
        self.mass_entry.setKeyboardTracking(False)

        mass_label = QtWidgets.QLabel('Sample mass (mg)')
        col_2_layout.addWidget(mass_label)
        col_2_layout.addWidget(self.mass_entry)

        # 3rd column - submit button
        col_3_widget = QtWidgets.QWidget(parent=self.widget)
        col_3_layout = QtWidgets.QHBoxLayout(col_3_widget)

        submit_button = QtWidgets.QPushButton(
            parent=col_3_widget,
            text='Submit'
        )
        submit_button.clicked.connect(
            lambda x: self.parse_info(user_cfg)
        )
        col_3_layout.addWidget(submit_button)

        layout.addWidget(col_1_widget)
        layout.addWidget(col_2_widget)
        layout.addWidget(col_3_widget)

        return

    def parse_info(self, user_cfg: utils.UserConfig):
        '''
        Callback function for qt window for mass and molecular weight

        Parameters
        ----------
        user_cfg: utils.UserConfig
            UserConfig object which will contain mass and mw

        Returns
        -------
        None
        '''

        mass = float(self.mass_entry.text())
        mw = float(self.mw_entry.text())

        if mass <= 0.:
            error_box('Mass cannot be <= zero')
        if mw <= 0.:
            error_box('MW cannot be <= zero')

        setattr(user_cfg, 'mass', mass)
        setattr(user_cfg, 'mw', mw)

        self.close_continue = True
        self.close()

        return

    def closeEvent(self, event):
        '''
        Catch if user closes prompt
        '''
        if self.close_continue:
            self.close()
        else:
            sys.exit()


def mass_mw_entry(user_cfg: utils.UserConfig) -> None:
    '''
    Creates qt window for user to input mass and molecular weight of
    sample

    Parameters
    ----------
    user_cfg: utils.UserConfig
        Configuration object

    Returns
    -------
    None
    '''

    app = QtWidgets.QApplication([])

    window = SampleInfoEntry(user_cfg)
    window.show()

    app.exec()

    del app

    return


def filename_entry(user_cfg: utils.UserConfig) -> None:
    '''
    Creates qt window for user to input mass and molecular weight of
    sample

    Parameters
    ----------
    user_cfg: utils.UserConfig
        Configuration object

    Returns
    -------
    None
    '''

    _ = QtWidgets.QApplication(sys.argv)

    window = QtWidgets.QMainWindow()

    file = QtWidgets.QFileDialog.getOpenFileName(
        window,
        caption="File to Import",
        directory=".",
        filter="Text files (*.dat *.txt *.csv)"
    )

    # Check for empty filename - user has closed window
    if file[0] is None:
        sys.exit()
    elif not len(file[0]):
        sys.exit()
    else:
        user_cfg.file_name = file[0]

    return


def make_mpl_radiobuttons(pos: list[float], labels: list[str],
                          figure: plt.Figure, circle_area: float = 100.,
                          ax: str | plt.Axes = 'create',
                          facecolor: str = 'blue') -> RadioButtons:
    '''
    Creates matplotlib radiobuttons on given figure

    Parameters
    ----------
    pos: list[float]
        Four values specifying left, bottom, width, height of
        rectangle in which buttons will reside
    labels: list[str]
        Label for each radiobutton
    figure: plt.Figure
        Figure to which radiobuttons are added
    circle_area: float, default 100.
        Area of radiobutton circle
    ax: str | plt.Axes, default 'create'
        If string 'create', a new axis is made, else the provided axis is used
    facecolor: str
        Name of colour used for button when activated

    Returns
    -------
    matplotlib.widgets.RadioButtons
        Set of radiobuttons for matplotlib figure window
    '''

    if ax == 'create':
        ax = plt.axes(
            pos,
            facecolor='w',
            frameon=False,
            aspect='equal',
            figure=figure
        )
    buttons = RadioButtons(
        ax,
        labels,
        radio_props={
            's': circle_area,
            'facecolor': facecolor
        }
    )

    if len(labels) == 1:
        fc = [0, 0, 0, 0]
    else:
        fc = [
            [0, 0, 0, 0] for _ in range(len(labels))
        ]

    buttons._buttons.set_facecolor(
        fc
    )

    return buttons


def create_ac_temp_colorbar(ax: plt.Axes, fig: plt.Figure,
                            temps: list[float],
                            colors: mpl_col.Colormap) -> Colorbar:
    '''
    Creates colorbar for temperatures in AC plotting

    Parameters
    ----------

    ax: plt.Axes
        Axis to which colorbar is added
    fig: plt.Figure
        Figure to which colorbar is added
    temps: list[float]
        Temperatures in Kelvin
    colors: matplotlib.colors.Colormap
        Colormap used in plot

    Returns
    -------
    matplotlib.colorbar.Colorbar
        Matplotlib colorbar object

    Raises
    ------
    ValueError
        If no temperatures specified
    '''

    n_temps = len(temps)

    if n_temps == 0:
        raise ValueError('Cannot create colorbar for zero temperatures')

    # Make colourbar
    # Indexing starts at zero and ends at num_temps
    norm = mpl_col.BoundaryNorm(
        np.arange(0, n_temps + 1),
        ncolors=colors.N
    )

    # Scalar mappable converts colourmap numbers into an image of colours
    sm = cm.ScalarMappable(cmap=colors, norm=norm)

    colorbar_ticks = np.arange(1, n_temps + 1) - 0.5
    colorbar_labels = get_temp_colourbar_ticks(temps)

    cbar = fig.colorbar(
        sm,
        ticks=colorbar_ticks,
        orientation='horizontal',
        format='%.1f',
        cax=ax
    )

    ax.set_xticklabels(
        colorbar_labels,
        rotation=0,
        fontsize='smaller'
    )

    ax.minorticks_off()

    # Set colourbar label - technically title - above bar
    cbar.ax.set_title('T (K)', fontsize='smaller')

    return cbar


def get_temp_colourbar_ticks(temperatures: list[float]) -> list[float]:
    '''
    Creates ticks for a temperature colourbar.

    If there are fewer than 9 data points, give tick to all temperatures

    If there are between 9 and 17, slice every 3 or 4 points depending on
    even-odd.

    If there are more than 17, slice every 5 or 6 points depending on
    even-odd.

    Parameters
    ----------
    temperatures: list[float]
        Temperatures in Kelvin

    Returns
    -------
    list[float]
        Colourbar tick positions
    '''

    n_temps = len(temperatures)

    ticks = ['{:.1f}'.format(tp) for tp in temperatures]

    if n_temps <= 8:
        step = 1
    elif 9 <= n_temps <= 17:
        if n_temps % 2 == 0:
            step = 3
        else:
            step = 4
    else:
        if n_temps % 2 == 0:
            step = n_temps // 5
        else:
            step = n_temps // 4

    # Swap numbers for blanks, and ensure start and end are present
    ticks = [ti if not it % step else '' for (it, ti) in enumerate(ticks)]
    ticks[0] = '{:.1f}'.format(temperatures[0])
    ticks[-1] = '{:.1f}'.format(temperatures[-1])

    # Remove adjacent labels at end
    if n_temps > 8 and ticks[-2] != '':
        ticks[-2] = ''

    return ticks


class SusceptibilityCanvas(FigureCanvasQTAgg):
    '''
    Figure and axes for AC Susceptibility plots
    '''

    def __init__(self, width, height, dpi=100, parent=None):
        self.fig = Figure(figsize=(width, height), dpi=dpi)
        self.gs = gridspec.GridSpec(3, 1, height_ratios=[0.05, 1, 1])
        self.ax = [
            self.fig.add_subplot(self.gs[0]),
            self.fig.add_subplot(self.gs[1]),
            self.fig.add_subplot(self.gs[2])
        ]
        super(SusceptibilityCanvas, self).__init__(self.fig)


class ColeColeCanvas(FigureCanvasQTAgg):
    '''
    Figure and axes for AC Cole-Cole plots
    '''

    def __init__(self, width, height, dpi=100, parent=None):
        self.fig = Figure(figsize=(width, height), dpi=dpi)
        self.gs = gridspec.GridSpec(2, 1, height_ratios=[0.03, 0.9])
        self.ax = [
            self.fig.add_subplot(self.gs[0]),
            self.fig.add_subplot(self.gs[1])
        ]
        super(ColeColeCanvas, self).__init__(self.fig)



'''This is the command line interface to cc-fit2'''
from . import ac, dc
from . import waveform as wfrm
from . import relaxation as relax
from . import utils as ut

import argparse
import numpy as np
import os
import sys
from qtpy import QtWidgets
import matplotlib.pyplot as plt
from glob import glob

import warnings
warnings.filterwarnings('ignore', '.*GUI is implemented.*')
warnings.filterwarnings('ignore', 'invalid value encountered in power')
warnings.filterwarnings('ignore', 'invalid value encountered in log10')
warnings.filterwarnings('ignore', 'invalid value encountered in divide')


# Set user specified font name
if os.getenv('ccfit2_fontname'):
    try:
        plt.rcParams['font.family'] = os.getenv('ccfit2_fontname')
    except ValueError:
        ut.cprint('Error in ccfit2_fontname environment variable', 'red')
        sys.exit(1)

# Set user specified font size
if os.getenv('ccfit2_fontsize'):
    try:
        plt.rcParams['font.size'] = float(os.getenv('ccfit2_fontsize'))
    except ValueError:
        ut.cprint('Error in ccfit2_fontsize environment variable', 'red')
        sys.exit(1)

# Set user specified plot file format
PFF = '.png'
if os.getenv('ccfit2_plot_format'):
    try:
        PFF = os.getenv('ccfit2_plot_format')
        if PFF[0] != '.':
            PFF = f'.{PFF}'
    except ValueError:
        ut.cprint('Error in ccfit2_plot_format environment variable', 'red')
        sys.exit(1)


_SHOW_CONV = {
    'on': True,
    'save': False,
    'show': True
}

_SAVE_CONV = {
    'on': True,
    'save': True,
    'show': False
}


def ac_mode_func(uargs):
    '''
    Wrapper function for command line interface call to ac mode

    Parameters
    ----------
    uargs : argparser object
        command line arguments

    Returns
    -------
    None
    '''

    # Make QT application - this MUST happen before any matplotlib/qt
    # plots are made
    app = QtWidgets.QApplication([])

    # Object to store user configuration, paths, etc.
    user_cfg = ut.UserConfig()
    user_cfg.file_name = uargs.input_file

    # mass, mw
    user_cfg.mass = uargs.mass
    user_cfg.mw = uargs.MW

    ut.cprint('\n Parsing Input File', 'black_bluebg')
    # Check data_pointer is in file
    data_line = ut.locate_data_header(
        user_cfg.file_name, data_header=uargs.data_header
    )

    # If data_pointer is not present, print error message
    if data_line < 0:
        message = '\n    ***Error*** \n'
        message += '    --data_header {}\n'.format(uargs.data_header)
        message += '    is not present in \n'
        message += '    {}\n'.format(user_cfg.file_name)
        ut.cprint(message, 'red')
        sys.exit(1)

    # Get file headers
    header_indices, header_names = ut.parse_headers(
        user_cfg.file_name, data_line, ac.HEADERS_SUPPORTED
    )

    # If any headers are not present, print error message
    if any(val == -1 for val in header_indices.values()):
        message = '\n***Error***:'
        for ind, name in zip(header_indices.values(), header_names.keys()):
            if ind == -1:
                message += '\n {} column not found.'.format(name.upper())
                message += '\n  Supported headers for this column are: '
                for i in ac.HEADERS_SUPPORTED[name]:
                    message += '\n   {} '.format(i)
        message += '\nManually change the input file or raise'
        message += ' a GitLab Issue\n'
        ut.cprint(message, 'red')
        sys.exit(1)

    # Load data from file as unordered list of individual Measurements
    ac_measurements = ac.Measurement.from_file(
        user_cfg.file_name, mass=user_cfg.mass, mw=user_cfg.mw,
        header_indices=header_indices, data_header=uargs.data_header,
    )

    # Group measurements into sublists of experiments
    # Where first dimension uses dc field strength
    # and second uses temperature
    all_experiments_sep = ac.Experiment.from_measurements(
        ac_measurements, temp_thresh=uargs.temp_thresh,
        field_thresh=uargs.field_thresh
    )

    if uargs.verbose:
        print('\nVerbose mode entered.\n')
        for experiments in all_experiments_sep:
            print('Field: {} Oe'.format(experiments[0].rep_dc_field))
            for experiment in experiments:
                print(
                    'The raw temperatures at this field have been grouped as:'
                )
                print(
                    'Mean={:04.1f}, Min={}, Max={}\nRaw={}'.format(
                        experiment.rep_temperature,
                        np.min(experiment.raw_temperatures),
                        np.max(experiment.raw_temperatures),
                        experiment.raw_temperatures
                    )
                )
                print('{} Frequencies: {}\n'.format(
                    len(experiment.ac_freqs), experiment.ac_freqs)
                )

    # Manual selection of temperatures to fit
    if uargs.select_T:
        all_experiments_sep = [
            ac.interactive_t_select(experiments)
            for experiments in all_experiments_sep
        ]

    elif uargs.unselect_T:
        all_experiments_sep = [
            ac.interactive_t_select(experiments, on=True)
            for experiments in all_experiments_sep
        ]

    if uargs.process == 'plot':
        for experiments in all_experiments_sep:
            # Base save names for plots
            save_name = f'{experiments[0].rep_dc_field:.1f}Oe{PFF}'

            cc_save_name = f'cole_cole_{save_name}'

            ac.plot_colecole(
                experiments,
                save=True,
                show=False,
                save_name=os.path.join(user_cfg.results_dir, cc_save_name),
                window_title='Cole-Cole plot for {:.1f} Oe'.format(
                    experiments[0].rep_dc_field
                )
            )

            sus_save_name = f'susc_{save_name}'

            ac.plot_susceptibility(
                experiments,
                save=True,
                show=False,
                save_name=os.path.join(user_cfg.results_dir, sus_save_name),
                window_title='AC susceptibility for {:.1f} Oe'.format(
                    experiments[0].rep_dc_field
                )
            )
            plt.show()
        sys.exit()

    unique_dc_fields = np.unique(
        [experiment[0].rep_dc_field for experiment in all_experiments_sep]
    )

    # For each AC experiment, fit an AC model to the data
    for f_it, (experiments, dc_field) in enumerate(zip(all_experiments_sep, unique_dc_fields)): # noqa

        _msg = '\n Fitting AC Susceptibility at {:.5f} Oe'.format(
            dc_field
        )
        ut.cprint(_msg, 'black_bluebg')

        # Interactive window to choose AC model
        # for current set of experiments
        chosen_model = ac.interactive_ac_model_select(
            experiments
        )

        # Create a model instance for each experiment
        # all parameters are fitted in cli mode
        fit_vars = {var: 'guess' for var in chosen_model.PARNAMES}
        fix_vars = {}
        models: list[ac.Model] = [
            chosen_model(
                fit_vars=fit_vars, fix_vars=fix_vars, experiment=experiment
            )
            for experiment in experiments
        ]
        # Update flat lines threshold
        for model in models:
            model.flat_thresh = uargs.flat_thresh

        # Feed fits forward in temperature
        # Use T=n fit for T=n+1 guess
        prev_fit = 'predefined'
        for model, experiment in zip(models, experiments):

            if uargs.verbose:
                print('For T= {:.4f} K and H= {:.4f} Oe'.format(
                    experiment.rep_temperature, experiment.rep_dc_field
                ))
                print('Fixed parameters are:')
                print(model.fix_vars)
                print('Fit parameter guesses are:')
                print(model.fit_vars)

            model.fit_to(
                experiment, guess=prev_fit, no_discard=uargs.discard_off
            )

            if uargs.verbose:
                print('Final parameters are:')
                print(model.final_var_values)

            # Use these fitted params as next guess, if fit successful
            if model.fit_status:
                prev_fit = model.final_var_values
            # Else use this model's guess as next guess
            else:
                prev_fit = model.fit_vars

        # For each field, save to file all associated fit params and model
        # funcs
        base_filename = 'ac_{:.1f}Oe_{}'.format(
            dc_field,
            chosen_model.NAME.lower().replace(' ', '_')
        )
        base_filename = os.path.join(user_cfg.results_dir, base_filename)
        fit_filename = '{}_params.out'.format(base_filename)
        model_filename = '{}_model.out'.format(base_filename)

        ac.write_model_params(models, fit_filename)
        ac.write_model_data(experiments, models, model_filename)

        if all(not model.fit_status for model in models):
            ut.cprint('\n    ***Error***:', 'red')
            ut.cprint(
                'At {:.5f} Oe, no peaks were measured.'.format(dc_field),
                'red'
            )
            # Exit if only one field
            if len(unique_dc_fields) == 1:
                sys.exit(1)
            else:
                continue

        # Fitted cole cole plot, either separate or on one figure
        if uargs.single_plots != 'off':
            for experiment, model in zip(experiments, models):
                single_cc_save_name = 'ac_cole_cole_{:.2f}K_{:.3}Oe_{}{}'.format( # noqa
                    experiment.rep_temperature,
                    experiment.rep_dc_field,
                    model.NAME.lower().replace(' ', '_'),
                    PFF
                )
                single_sus_save_name = 'ac_susc_{:.2f}K_{:.3}Oe_{}{}'.format(
                    experiment.rep_temperature,
                    experiment.rep_dc_field,
                    model.NAME.lower().replace(' ', '_'),
                    PFF
                )
                ac.plot_single_fitted_cole_cole(
                    experiment,
                    model,
                    save_name=os.path.join(
                        user_cfg.results_dir,
                        single_cc_save_name
                    ),
                    save=uargs.single_plots in ['save', 'on'],
                    show=False
                )
                ac.plot_single_fitted_susceptibility(
                    experiment,
                    model,
                    save_name=os.path.join(
                        user_cfg.results_dir,
                        single_sus_save_name
                    ),
                    save=uargs.single_plots in ['save', 'on'],
                    show=False
                )
                if uargs.single_plots in ['show', 'on']:
                    plt.show()
                plt.close('all')

        cc_save_name = f'{base_filename}_cole_cole{PFF}'
        ac.plot_fitted_colecole(
            experiments,
            models,
            save_name=os.path.join(user_cfg.results_dir, cc_save_name),
            save=True,
            show=False
        )

        sus_save_name = f'{base_filename}_susc{PFF}'
        ac.plot_fitted_susceptibility(
            experiments,
            models,
            save_name=os.path.join(user_cfg.results_dir, sus_save_name),
            save=True,
            show=False
        )
        plt.show()

        # If only susc, loop back here
        if uargs.process == 'susc':
            continue

        # Proceed to fit the relaxation profile after having finished fitting
        # the AC data.
        _msg = '\n Fitting relaxation rate vs temperature at {:.2f} Oe\n'.format( # noqa
            dc_field
        )
        ut.cprint(_msg, 'black_bluebg')

        if np.sum([model.fit_status for model in models]) <= 2:
            ut.cprint(
                '    ***Warning***:\n    Not enough data points to fit the relaxation profile', # noqa
                'red'
            )
            continue

        # Create dataset to store relaxation rate and temperature data
        datasets = relax.TauTDataset.from_ac_dc(models)

        for d_it, dataset in enumerate(datasets):

            # Ask user to select relaxation processes to fit, select
            # initial values, and fix parameters if desired

            rmodels, fit_vars, fix_vars, exited = relax.interactive_fitting(
                dataset, app
            )

            if exited:
                if len(unique_dc_fields) > 1 and f_it != len(unique_dc_fields) - 1: # noqa
                    ut.cprint('\n Skipping to next field', 'black_yellowbg')
                    continue
                elif len(datasets) > 1 and d_it != len(datasets) - 1:
                    ut.cprint('\n Skipping to next dataset', 'black_yellowbg')
                    continue
                else:
                    ut.cprint('\n Exiting fit window', 'red')
                    sys.exit(1)

            if not any(len(fv) for fv in fit_vars):
                ut.cprint('\n Error: All parameters fixed', 'red')
                sys.exit(1)

            # Create MultiLogModel as combination of individual models
            multilogmodel = relax.MultiLogTauTModel(
                rmodels,
                fit_vars,
                fix_vars
            )

            # Fit to experiment
            multilogmodel.fit_to(dataset)

            model_names = ''
            for _mod in multilogmodel.logmodels:
                model_names += '{}_'.format(_mod.NAME.lower())

            # Plot data
            base_filename = f'relaxation_{dc_field:.1f}Oe_{model_names}'
            # add number if more than one dataset is fitted
            if len(datasets) > 1:
                base_filename = '{}_set_{:d}'.format(base_filename, d_it + 1)

            base_filename = os.path.join(user_cfg.results_dir, base_filename)

            fit_plot_filename = f'{base_filename}fitted{PFF}'
            residuals_plot_filename = f'{base_filename}residuals{PFF}'

            relax.plot_fitted_rates(
                dataset,
                multilogmodel,
                save=True,
                show=True,
                save_name=fit_plot_filename
            )

            relax.plot_rate_residuals(
                dataset,
                multilogmodel,
                save=True,
                show=True,
                save_name=residuals_plot_filename
            )

            # Save raw data
            fit_filename = '{}params.out'.format(base_filename)
            model_filename = '{}model.out'.format(base_filename)

            relax.write_model_params(multilogmodel, fit_filename)
            relax.write_model_data(dataset, multilogmodel, model_filename)

    return


class FitFixAction(argparse.Action):
    '''
    Custom argparse action for Fit and Fix variable optional arguments
    '''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __call__(self, parser, namespace, values: list[str], option_string):
        '''
        When called by argparse, checks the fit and fit variables have
        two entries, the first is either string 'fit' or 'fix'
        and second is either:
        string 'guess'
        a float
        or string containing filename of parameters file
        '''
        if values[0].lower() not in ['fit', 'fix']:
            parser.error('First option must be fit or fix')
        elif ut.can_float(values[1]):
            values[1] = float(values[1])
        elif isinstance(values[1], str):
            # Check file exists
            if values[1].lower() == 'guess':
                pass
            elif not os.path.exists(values[1]):
                parser.error(
                    'Second option must be either: the word \'guess\', a filename, or a value' # noqa
                )

        # Add True to signify that the user added this argument to their cli
        # call
        values.append(True)

        setattr(namespace, self.dest, values)


class MultiAction(argparse.Action):
    '''
    Custom argparse action for multiple variable optional arguments
    '''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __call__(self, parser, namespace, values: list[str], option_string):
        '''
        When called by argparse, checks the variables have
        three entries, the first two are strings containing filenames
        and the third is either:
        a float
        or string containing filename of parameters file
        '''
        if not os.path.exists(values[0]):
            parser.error(
                'First option must be a filename'
            )
        elif not os.path.exists(values[1]):
            parser.error(
                'Second option must be a filename'
            )
        elif ut.can_float(values[2]):
            values[2] = float(values[2])
        else:
            parser.error(
                'Third option must be a float'
            )

        setattr(namespace, self.dest, values)


class DiffAction(argparse.Action):
    '''
    Custom argparse action for optional arguments which are can be floats or
    filenames
    '''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __call__(self, parser, namespace, value, option_string):
        '''
        When called by argparse, checks the variables have
        one entry, either:
        a float, or string containing filename of parameters file
        '''

        if ut.can_float(value):
            value = float(value)
        elif isinstance(value, str):
            # Check file exists
            if not os.path.exists(value):
                parser.error(
                    'Second option must be either: a value or a filename'
                )

        setattr(namespace, self.dest, value)


def extract_dc_param_args(arg_val: str | float,
                          n_rep: int) -> list[str | float]:
    '''
    Converts input fit/fix value option into list of n_rep entries containing
    values as 'guess', a float, or floats read from file

    Parameters
    ----------
    arg_val: str | float
        Fit/Fix value, either string 'guess', a float, or a filename string
    n_rep: int
        Number of times to repeat 'guess' or float

    Returns
    -------
    list[str | float]
        list of n_rep entries containing values as 'guess', a float,
        or floats read from file
    '''

    if arg_val == 'guess':
        output = ['guess'] * n_rep
    elif isinstance(arg_val, float):
        output = [arg_val] * n_rep
    else:
        encoding = ut.detect_encoding(arg_val)
        output = np.loadtxt(arg_val, skiprows=1, usecols=0, encoding=encoding)
        if len(output) != n_rep:
            ut.cprint(
                f'Error: File {arg_val} length does not match number of experiments', # noqa 
                'red'
            )
            sys.exit(1)

    return output


def extract_dc_field_calibration_args(arg_val: str | float,
                                      n_rep: int) -> list[str | float]:
    '''
    Converts input MultiAction value option into list of n_rep entries
    containing values as a float or floats read from file

    Parameters
    ----------
    arg_val: str | float
        Fit/Fix value, either a float or a filename string
    n_rep: int
        Number of times to repeat float

    Returns
    -------
    list[str | float]
        list of n_rep entries containing values as a float,
        or floats read from file
    '''

    if isinstance(arg_val, float):
        output = [arg_val] * n_rep
    else:
        encoding = ut.detect_encoding(arg_val)
        output = np.loadtxt(arg_val, skiprows=0, usecols=0, encoding=encoding)
        if len(output) != n_rep:
            ut.cprint(
                f'Error: File {arg_val} length does not match number of experiments', # noqa 
                'red'
            )
            sys.exit(1)

    return output


def calibrate_meq(sat_file: str, calc_file: str, sat_field: float,
                  n_exp: int) -> list[str | float]:
    '''
    Calibrates M_eq using theoretically calculated equilibrium magnetisation
    data as a function of field and the experimentally determined magnetisation
    at saturation

    Parameters
    ----------
    sat_file: str
        Name of file containing experimental saturation fields
    calc_file: str
        Name of file containing theoretical magnetisation curve
    sat_field: float
        Value for saturation field for all experiments.
    n_exp: int
        Number of experiments to expect in file

    Returns
    -------
    list[float]
        Calibrated values of M_eq, one per experiment
    '''

    encoding0 = ut.detect_encoding(sat_file)
    encoding1 = ut.detect_encoding(calc_file)
    sat_mag_data = np.loadtxt(sat_file, skiprows=0, encoding=encoding0)
    calc_mag_data = np.loadtxt(calc_file, skiprows=0, encoding=encoding1)

    if len(sat_mag_data) != n_exp:
        ut.cprint(
            f'Error: File {sat_file} length does not match number of experiments', # noqa 
            'red'
        )
        sys.exit(1)

    # Pick out calculated moment at experimental saturation field
    calc_sat_pos = np.where(calc_mag_data == sat_field)
    calc_sat = calc_mag_data[calc_sat_pos[0], 1][0]

    # create empty list to fill with M_eq values
    calib_Meq = []
    for i, sats in enumerate(sat_mag_data[:, 0]):
        # copy theoretically calculated Meq
        calib_mag_data = np.copy(calc_mag_data)

        # Calculate calibration factor
        # as ratio of calculated moment verses experimental moment at
        # experimental saturation field
        factor = calc_sat / sat_mag_data[i, 1]

        # apply calibration factor to experimental moment
        calib_mag_data[:, 1] /= factor

        # pick out calibrated M_eq for each experiment
        calib_Meq_pos = np.where(calib_mag_data == sats)
        calib_Meq.append(calib_mag_data[calib_Meq_pos[0], 1][0])

    # Catch empty list of calibrated values
    if not len(calib_Meq):
        ut.cprint(
            f'Error: M_eq calibration failed. Please check input', # noqa 
            'red'
        )
        sys.exit(1)

    return calib_Meq


def dc_mode_func(uargs):
    '''
    Wrapper function for command line interface call to dc mode

    Parameters
    ----------
    uargs : argparser object
        command line arguments

    Returns
    -------
    None
    '''

    # Parse model parameter arguments to enforce mutually exclusive groups
    stretched_par_args = [uargs.taustar, uargs.beta]
    double_par_args = [
        uargs.taustar1, uargs.beta1, uargs.taustar2,
        uargs.beta2, uargs.frac
    ]

    # Check if double parameters specified with stretched model
    if uargs.model == 'stretched':
        if any(len(par_arg) == 3 for par_arg in double_par_args):
            ut.cprint(
                'Error: Specified parameters do not match --model stretched',
                'red'
            )
            sys.exit(1)

    # Check if stretched parameters specified with double model
    if uargs.model == 'double':
        if any(len(par_arg) == 3 for par_arg in stretched_par_args):
            ut.cprint(
                'Error: Specified parameters do not match --model double',
                'red'
            )
            sys.exit(1)

    # Make QT application - this MUST happen before any matplotlib/qt
    # plots are made
    app = QtWidgets.QApplication([])

    # Object to store user configuration, paths, etc.
    user_cfg = ut.UserConfig()
    user_cfg.file_name = uargs.input_file

    all_measurements = []
    ut.cprint('\n Parsing Input Files', 'black_bluebg')

    # Use glob to expand wildcards if on windows
    if 'nt' in os.name and '*' in user_cfg.file_name[0]:
        user_cfg.file_name = glob(user_cfg.file_name[0])

    for fname in user_cfg.file_name:

        # Check data_pointer is in file
        data_line = ut.locate_data_header(
            fname, data_header=uargs.data_header
        )

        # If data_pointer is not present, print error message
        if data_line < 0:
            message = '\n    ***Error*** \n'
            message += '    --data_header {}\n'.format(uargs.data_header)
            message += '    is not present in \n'
            message += '    {}\n'.format(fname)
            ut.cprint(message, 'red')
            sys.exit(1)

        # Get file header indices and names
        # enforce specific moment header if user requests
        header_dict = dc.HEADERS_SUPPORTED
        if uargs.moment_header == 'moment':
            header_dict['moment'] = ['Moment (emu)']
        if uargs.moment_header == 'fixed':
            header_dict['moment'] = ['DC Moment Fixed Ctr (emu)']
        if uargs.moment_header == 'free':
            header_dict['moment'] = ['DC Moment Free Ctr (emu)']

        header_indices, header_names = ut.parse_headers(
            fname, data_line, header_dict
        )

        if uargs.verbose:
            print(f'Using the following headers in file {fname}')
            print(header_names)

        # If any headers are not present, or are empty print error message
        if any(val == -1 for val in header_indices.values()):
            message = '\n***Error***:'
            for ind, name in zip(header_indices.values(), header_names.keys()):
                if ind == -1:
                    message += '\n {} column not found.'.format(name.upper())
                    message += '\n  Supported headers for this column are: '
                    for i in header_dict[name]:
                        message += '\n   {} '.format(i)
                elif ind == -2:
                    message += '\n {} column is empty.'.format(name.upper())
            message += '\nManually change the input file or raise'
            message += ' a GitLab Issue\n'
            ut.cprint(message, 'red')
            sys.exit(1)

        # Read current file and extract data from columns to create
        # measurement, one per row
        measurements = dc.Measurement.from_file(
            fname, header_indices, uargs.data_header
        )

        # Trim measurements to ignore values at saturating and
        # decaying field
        if not uargs.no_field_discard:
            diff_thresh = uargs.dfield_thresh
            # Forward difference of field
            field_diff = np.abs(
                np.diff([measurement.dc_field for measurement in measurements])
            )
            # Flip field diff and find first value > threshold, then index to
            # unflipped
            n_pts = len(measurements)
            cut_at = n_pts - np.argmax(np.flip(field_diff) > diff_thresh) - 1

            # Cut to stable field values only
            measurements = measurements[cut_at:]

        # Add to list of all measurements
        all_measurements += measurements

    ut.cprint('.... Done!', 'black_bluebg')

    # Group measurements into lists of experiments
    # Using dc field and temperature
    experiments = dc.Experiment.from_measurements(
        all_measurements,
        temp_thresh=uargs.temp_thresh,
        cut_moment=uargs.cut_moment
    )

    # convert measured dc field values to calibrated field values if called for
    if uargs.field_calibration is not None:
        actual_fields = extract_dc_field_calibration_args(
            uargs.field_calibration, len(experiments)
        )
        for experiment, field in zip(experiments, actual_fields):
            experiment.dc_fields[:] = field

    # Plot moment and dc field against time and exit if requested
    if uargs.process == 'plot':
        file_name = f'raw_decays{PFF}'
        file_name = os.path.join(user_cfg.results_dir, file_name)
        dc.plot_decays_and_fields(
            experiments, show=True, save=True,
            x_scale=uargs.plot_axes[0], y_scale=uargs.plot_axes[1],
            save_name=file_name
        )
        sys.exit()

    # Fit decays to exponential function

    # Create models, one per experiment
    # specifying type of model, and which parameters will
    # be fitted and which are fixed
    models: list[dc.Model] = []

    if uargs.model == 'stretched':

        parargs = [
            uargs.taustar, uargs.beta, uargs.M_eq,
            uargs.M_0, uargs.t_offset
        ]

        names = [
            'tau*', 'beta', 'm_eq', 'm_0', 't_offset'
        ]

        # Convert each fit parameter into a list with n_input_files entries
        all_vars = {
            name: [
                arg[0], extract_dc_param_args(arg[1], len(experiments))
            ]
            for name, arg in zip(names, parargs)
        }

        all_fit_vars = {
            name: value[1]
            for name, value in all_vars.items()
            if 'fit' in value[0]
        }

        all_fix_vars = {
            name: value[1]
            for name, value in all_vars.items()
            if 'fix' in value[0]
        }

        # Calibrate M_eq based on theoretical values
        if uargs.M_eq_calibration is not None:
            all_fix_vars['m_eq'] = calibrate_meq(
                *uargs.M_eq_calibration,
                len(experiments)
            )

        model_to_use = dc.ExponentialModel
        for it, experiment in enumerate(experiments):
            fit_vars = {
                key: all_fit_vars[key][it]
                for key in all_fit_vars.keys()
            }
            fix_vars = {
                key: all_fix_vars[key][it]
                for key in all_fix_vars.keys()
            }
            models.append(model_to_use(fit_vars, fix_vars, experiment))

    elif uargs.model == 'double':

        parargs = [
            uargs.taustar1, uargs.taustar2, uargs.beta1,
            uargs.beta2, uargs.M_eq,
            uargs.M_0, uargs.t_offset, uargs.frac
        ]

        names = [
            'tau*1', 'tau*2', 'beta1', 'beta2', 'm_eq', 'm_0', 't_offset',
            'frac'
        ]

        # Convert each fit parameter into a list with n_input_files entries
        all_vars = {
            name: [
                arg[0], extract_dc_param_args(arg[1], len(experiments))
            ]
            for name, arg in zip(names, parargs)
        }

        all_fit_vars = {
            name: value[1]
            for name, value in all_vars.items()
            if 'fit' in value[0]
        }

        all_fix_vars = {
            name: value[1]
            for name, value in all_vars.items()
            if 'fix' in value[0]
        }

        # Calibrate M_eq based on theoretical values
        if uargs.M_eq_calibration is not None:
            all_fix_vars['m_eq'] = calibrate_meq(
                *uargs.M_eq_calibration,
                len(experiments)
            )
        model_to_use = dc.DoubleExponentialModel
        for it, experiment in enumerate(experiments):
            fit_vars = {
                key: all_fit_vars[key][it]
                for key in all_fit_vars.keys()
            }
            fix_vars = {
                key: all_fix_vars[key][it]
                for key in all_fix_vars.keys()
            }

            models.append(model_to_use(fit_vars, fix_vars, experiment))

    ut.cprint('\n Fitting DC Decays\n', 'black_bluebg')
    # For each model and accompanying experiment, fit model
    for it, (model, experiment) in enumerate(zip(models, experiments)):

        if uargs.verbose:
            print('Fit parameter guesses are:')
            print(model.fit_vars)
            print('Fixed parameters are:')
            print(model.fix_vars)
            print()
        # Fit model
        model.fit_to(experiment)

        # Plot decay with fit
        if model.fit_status:
            if uargs.decay_plots in ['on', 'show', 'save']:
                file_name = 'fitted_decay_{:.2f}K_{:.1f}Oe{}'.format(
                    experiment.rep_temperature,
                    experiment.dc_fields[-1],
                    PFF
                )

                file_name = os.path.join(user_cfg.results_dir, file_name)
                dc.plot_fitted_decay(
                    experiment=experiment, model=model,
                    show=_SHOW_CONV[uargs.decay_plots],
                    save=_SAVE_CONV[uargs.decay_plots],
                    save_name=file_name, x_scale=uargs.plot_axes[0],
                    y_scale=uargs.plot_axes[1],
                    show_params=uargs.hide_params
                )
        else:
            ut.cprint(
                        '\n Cannot save/show plot fot T = {} K and {} Oe as fit has failed'.format( # noqa
                            model.temperature, model.dc_field
                        ),
                        'black_yellowbg'
                )

    base_filename = 'dc_{}_'.format(model.NAME.lower().replace(' ', '_'))
    base_filename = os.path.join(user_cfg.results_dir, base_filename)

    # Save parameters to file
    dc.write_model_params(
        models,
        file_name='{}params.out'.format(base_filename)
    )
    # Save modelled data to file
    dc.write_model_data(
        experiments,
        models,
        file_name='{}model.out'.format(base_filename)
    )
    # If only one temperature for all points, then quit
    if len(np.unique([e.rep_temperature for e in experiments])) == 1:
        sys.exit()

    # exit here if user requests it
    if uargs.process == 'decays':
        sys.exit()
    ut.cprint('\n Fitting relaxation rate vs temperature\n', 'black_bluebg')

    # Create dataset to store relaxation rate and temperature data
    datasets = relax.TauTDataset.from_ac_dc(models)

    for d_it, dataset in enumerate(datasets):

        # Ask user to select relaxation processes to fit, select
        # initial values, and fix parameters if desired
        rmodels, fit_vars, fix_vars, exited = relax.interactive_fitting(
            dataset, app
        )

        if exited:
            ut.cprint('\n Exiting fit window', 'red')
            sys.exit(1)

        if not any(len(fv) for fv in fit_vars):
            ut.cprint('\n Error: All parameters fixed', 'red')
            sys.exit(1)

        # Create MultiLogModel as combination of individual models
        multilogmodel = relax.MultiLogTauTModel(
            rmodels,
            fit_vars,
            fix_vars
        )

        # Fit model to dataset
        multilogmodel.fit_to(dataset)

        model_names = ''
        for _mod in multilogmodel.logmodels:
            model_names += '{}_'.format(_mod.NAME.lower())

        # Plot data
        base_filename = f'relaxation_{model_names}'
        # add number if more than one dataset is fitted
        if len(datasets) > 1:
            base_filename = '{}_set_{:d}'.format(base_filename, d_it + 1)
        base_filename = os.path.join(user_cfg.results_dir, base_filename)

        fit_plot_filename = f'{base_filename}fitted{PFF}'
        residuals_plot_filename = f'{base_filename}residuals{PFF}'

        # Plot fitted model and experimental rates
        relax.plot_fitted_rates(
            dataset,
            multilogmodel,
            save=True,
            save_name=fit_plot_filename
        )

        # Plot log10(rate) residuals between fitted model and experimental
        # rates
        relax.plot_rate_residuals(
            dataset,
            multilogmodel,
            save=True,
            save_name=residuals_plot_filename
        )

        fit_filename = '{}params.out'.format(base_filename)
        model_filename = '{}model.out'.format(base_filename)

        relax.write_model_params(multilogmodel, fit_filename)
        relax.write_model_data(dataset, multilogmodel, model_filename)

    return


def relaxation_mode_func(uargs):
    '''
    Wrapper function for command line interface call to relaxation mode

    Parameters
    ----------
    uargs : argparser object
        command line arguments

    Returns
    -------
    None

    '''

    # Make QT application - this MUST happen before any matplotlib/qt
    # plots are made
    app = QtWidgets.QApplication([])

    # Use glob to expand wildcards if on windows
    if 'nt' in os.name and '*' in uargs.input_files[0]:
        uargs.input_files = glob(uargs.input_files[0])

    if uargs.x_var.lower() == 't':
        ut.cprint(
            '\n Fitting relaxation rate vs temperature\n',
            'black_bluebg'
        )

        if uargs.filetype == 'rate':
            dataset = relax.TauTDataset.from_rate_files(uargs.input_files)
        else:
            # Create dataset to store relaxation rate and temperature data
            dataset = relax.TauTDataset.from_ccfit2_files(
                uargs.input_files
            )

        # Plot rates and exit
        if uargs.process == 'plot':
            if uargs.plot_type in ['rate', 'both']:
                relax.plot_rates(
                    dataset,
                    save=True,
                    save_name=f'relaxation_rates{PFF}'
                )
            if uargs.plot_type in ['time', 'both']:
                relax.plot_times(
                    dataset,
                    save=True,
                    save_name=f'relaxation_times{PFF}'
                )
            sys.exit()

        # Disable weighting of residuals by setting uncertainties
        # as empty arrays
        if uargs.no_weights:
            dataset.lograte_pm = []

        # Ask user to select relaxation processes to fit, select
        # initial values, and fix parameters if desired
        rmodels, fit_vars, fix_vars, exited = relax.interactive_fitting(
            dataset, app
        )

        if exited:
            ut.cprint('\n Exiting fit window', 'red')
            sys.exit(1)

        if not any(len(fv) for fv in fit_vars):
            ut.cprint('\n Error: All parameters fixed', 'red')
            sys.exit(1)

        # Create MultiLogModel as combination of individual models
        multilogmodel = relax.MultiLogTauTModel(
            rmodels,
            fit_vars,
            fix_vars
        )

        # Fit to experiment
        multilogmodel.fit_to(dataset)

        # Construct string from model names
        model_names = ''
        for _mod in multilogmodel.logmodels:
            model_names += '{}_'.format(_mod.NAME.lower())

        # Plot fitted model and experiment
        if uargs.plot_type in ['rate', 'both']:
            relax.plot_fitted_rates(
                dataset,
                multilogmodel,
                save=True,
                save_name=f'relaxation_{model_names}fitted_rates{PFF}',
                show_params=uargs.hide_params
            )
        if uargs.plot_type in ['time', 'both']:
            relax.plot_fitted_times(
                dataset,
                multilogmodel,
                save=True,
                save_name=f'relaxation_{model_names}fitted_times{PFF}',
                show_params=uargs.hide_params
            )
        # Plot log10(rate) residuals between dataset and model
        relax.plot_rate_residuals(
            dataset,
            multilogmodel,
            save=True,
            save_name=f'relaxation_{model_names}residuals{PFF}'
        )

        fit_filename = f'relaxation_{model_names}params.out'
        model_filename = f'relaxation_{model_names}model.out'

        relax.write_model_params(multilogmodel, fit_filename)
        relax.write_model_data(dataset, multilogmodel, model_filename)

    elif uargs.x_var.lower() == 'h':
        ut.cprint('\n Fitting relaxation rate vs field\n', 'black_bluebg')

        if uargs.filetype == 'rate':
            dataset = relax.TauHDataset.from_rate_files(uargs.input_files)
        else:
            # Create dataset to store relaxation rate and temperature data
            dataset = relax.TauHDataset.from_ccfit2_files(
                uargs.input_files
            )

        # Plot rates and exit
        if uargs.process == 'plot':
            if uargs.plot_type in ['rate', 'both']:
                relax.plot_rates(
                    dataset,
                    save=True,
                    save_name=f'relaxation_rates{PFF}'
                )
            if uargs.plot_type in ['time', 'both']:
                relax.plot_times(
                    dataset,
                    save=True,
                    save_name=f'relaxation_times{PFF}'
                )
            sys.exit()

        if uargs.no_weights:
            dataset.lograte_pm = []
            dataset.rate_pm = []

        model_opt = [
            relax.LogFDQTMModel,
            relax.LogRamanIIModel,
            relax.LogConstantModel,
            relax.LogBVVRamanIIModel,
            relax.LogBVVConstantModel
        ]

        # Ask user to select relaxation processes to fit, select
        # initial values, and fix parameters if desired
        rmodels, fit_vars, fix_vars, exited = relax.interactive_fitting(
            dataset, app, model_opt=model_opt
        )

        if exited:
            ut.cprint('\n Exiting fit window', 'red')
            sys.exit(1)

        if not any(len(fv) for fv in fit_vars):
            ut.cprint('\n Error: All parameters fixed', 'red')
            sys.exit(1)

        # Create MultiLogModel as combination of individual models
        multilogmodel = relax.MultiLogTauHModel(
            rmodels,
            fit_vars,
            fix_vars
        )

        # Fit to experiment
        multilogmodel.fit_to(dataset)

        # Construct string from model names
        model_names = ''
        for _mod in multilogmodel.logmodels:
            model_names += '{}_'.format(_mod.NAME.lower().replace(' * ', '-'))

        # Plot fitted model and experiment
        if uargs.plot_type in ['rate', 'both']:
            relax.plot_fitted_rates(
                dataset,
                multilogmodel,
                save=True,
                save_name=f'relaxation_{model_names}fitted_rates{PFF}',
                show_params=uargs.hide_params
            )
        if uargs.plot_type in ['time', 'both']:
            relax.plot_fitted_times(
                dataset,
                multilogmodel,
                save=True,
                save_name=f'relaxation_{model_names}fitted_times{PFF}',
                show_params=uargs.hide_params
            )

        # Plot log10(rate) residuals between dataset and model
        relax.plot_rate_residuals(
            dataset,
            multilogmodel,
            save=True,
            save_name=f'relaxation_{model_names}residuals{PFF}'
        )

        fit_filename = f'relaxation_{model_names}params.out'
        model_filename = f'relaxation_{model_names}model.out'

        relax.write_model_params(multilogmodel, fit_filename)
        relax.write_model_data(dataset, multilogmodel, model_filename)

    return


def waveform_mode_func(uargs):
    '''
    Wrapper function for command line interface call to waveform mode

    Parameters
    ----------
    uargs : argparser object
        command line arguments

    Returns
    -------
    None

    '''

    user_cfg = ut.UserConfig()
    user_cfg.file_name = uargs.input_file

    # Use glob to expand wildcards if on windows
    if 'nt' in os.name and '*' in user_cfg.file_name[0]:
        user_cfg.file_name = glob(user_cfg.file_name[0])

    experiments: list[wfrm.Experiment] = []

    for fname in user_cfg.file_name:

        # Check data_pointer is in file
        data_line = ut.locate_data_header(
            fname, data_header=uargs.data_header
        )

        # If data_pointer is not present, print error message
        if data_line < 0:
            message = '\n    ***Error*** \n'
            message += '    --data_header {}\n'.format(uargs.data_header)
            message += '    is not present in \n'
            message += '    {}\n'.format(fname)
            ut.cprint(message, 'red')
            sys.exit(1)

        header_dict = wfrm.HEADERS_SUPPORTED

        header_indices, header_names = ut.parse_headers(
            fname, data_line, header_dict
        )

        # If any headers are not present, or are empty print error message
        if any(val == -1 for val in header_indices.values()):
            message = '\n***Error***:'
            for ind, name in zip(header_indices.values(), header_names.keys()):
                if ind == -1:
                    message += '\n {} column not found.'.format(name.upper())
                    message += '\n  Supported headers for this column are: '
                    for i in header_dict[name]:
                        message += '\n   {} '.format(i)
                elif ind == -2:
                    message += '\n {} column is empty.'.format(name.upper())
            message += '\nManually change the input file or raise'
            message += ' a GitLab Issue\n'
            ut.cprint(message, 'red')
            sys.exit(1)

    # Read each file as individual measurements
    measurements = [
        wfrm.Measurement.from_file(file, data_header=uargs.data_header)
        for file in user_cfg.file_name
    ]

    # Collapse into a single list
    measurements = [
        m
        for mm in measurements
        for m in mm
    ]

    # Group by DC field oscillation period
    all_experiments = wfrm.Experiment.from_measurements(
        measurements,
        uargs.field_window,
        uargs.temp_thresh
    )

    ac_experiments = []

    # Fourier transform each temperature's set of experiments
    # and create corresponding AC Experiment to store AC data
    for experiments in all_experiments:
        # Carry out Fourier transform of moment and field
        ft_results = [
            wfrm.FTResult.from_experiment(experiment)
            for experiment in experiments
        ]

        # Plot fourier transform if requested
        if uargs.ft_plots in ['on', 'show', 'save']:
            for ftr in ft_results:
                _save_name = 'waveform_FT_{:.2f}_mHz_{:.2f}_K{}'.format(
                    1 / ftr.period * 1000,
                    experiments[0].rep_temperature,
                    PFF
                )
                wfrm.plot_waveform_data(
                    ftr,
                    show=_SHOW_CONV[uargs.ft_plots],
                    save=_SAVE_CONV[uargs.ft_plots],
                    save_name=os.path.join(
                        user_cfg.results_dir,
                        _save_name
                    ),
                    window_title='Fourier Transformed Data for {:.2f} Hz'.format( # noqa
                        1 / ftr.period
                    ),
                    verbose=True
                )

        # Create ac.Experiment object from Fourier transform results
        ac_experiments.append(
            wfrm.FTResult.create_ac_experiment(
                ft_results, experiments, mass=None, mw=None
            )
        )

    if uargs.susc_plot in ['on', 'show', 'save']:
        # Plot susceptibility data
        _save_name = f'waveform_susc{PFF}'
        ac.plot_susceptibility(
            ac_experiments,
            show=_SHOW_CONV[uargs.susc_plot],
            save=_SAVE_CONV[uargs.susc_plot],
            save_name=os.path.join(user_cfg.results_dir, _save_name),
            verbose=True
        )

    # Create magnetometer file style input for ccfit2 ac mode
    _file_name = os.path.join(
        user_cfg.results_dir, 'waveform_ccfit2_ac_input.out'
    )

    ac.save_ac_magnetometer_file(
        ac_experiments,
        file_name=_file_name,
        verbose=False
    )

    ut.cprint(' Use this file as input to ccfit2 ac:', 'green')
    ut.cprint(f' {_file_name}', 'blue')

    return


def read_args(arg_list=None):
    '''
    Parser for command line arguments. Uses subparsers for individual programs

    Parameters
    ----------
    args : argparser object
        command line arguments

    Returns
    -------
    None

    '''

    description = '''
    Program to extract the relaxation times from AC or DC data and fit their
    relaxation profile.

    Available modules:
        ccfit2 ac ...
        ccfit2 waveform ...
        ccfit2 dc ...
        ccfit2 relaxation ...
    '''

    epilog = '''
    To display options for a specific module, use ccfit2 module -h
    '''
    for i, arg in enumerate(sys.argv):
        if (arg[0] == '-') and arg[1].isdigit():
            sys.argv[i] = f' {arg}'

    parser = argparse.ArgumentParser(
        description=description,
        epilog=epilog,
        formatter_class=argparse.RawDescriptionHelpFormatter
    )

    # create the top-level parser
    subparsers = parser.add_subparsers(dest='exe_mode')

    # AC mode
    description_ac = '''
    Extract relaxation times from AC susceptibilities using the Debye model and
    (optional) fit the resulting relaxation profile.
    '''

    ac_parser = subparsers.add_parser(
        'ac',
        description=description_ac,
        formatter_class=argparse.RawTextHelpFormatter
    )

    ac_parser.set_defaults(func=ac_mode_func)

    ac_parser.add_argument(
        'input_file',
        type=str,
        help='''File containing the AC raw data.
        See Manual for expected format'''
    )
    ac_parser.add_argument(
        'mass',
        type=float,
        help='Sample mass (mg)'
    )
    ac_parser.add_argument(
        'MW',
        type=float,
        help='Sample molecular weight (g/mol)'
    )
    ac_parser.add_argument(
        '--process',
        metavar='<Option>',
        choices=['plot', 'susc', 'all'],
        default='all',
        help=(
            'What to do:\n'
            ' - \'plot\' just shows the raw data\n'
            ' - \'susc\' fits only the AC data\n'
            ' - \'all\' fits AC data and relaxation profile\n'
            'Default: all'
        )
    )
    ac_parser.add_argument(
        '--discard_off',
        action='store_true',
        help='Fit the susceptibilities even if no peak can be found'
    )
    ac_parser.add_argument(
        '--temp_thresh',
        metavar='<float>',
        type=float,
        default=0.1,
        help='Threshold used to discriminate between temperatures, default 0.1 K' # noqa
    )
    ac_parser.add_argument(
        '--field_thresh',
        metavar='<float>',
        type=float,
        default=1,
        help='Threshold used to discriminate between DC Fields, default 1 Oe' # noqa
    )

    t_selector = ac_parser.add_mutually_exclusive_group()
    t_selector.add_argument(
        '--select_T',
        action='store_true',
        help=(
            'Interactively select the temperatures to fit\n'
        )
    )
    t_selector.add_argument(
        '--unselect_T',
        action='store_true',
        help=(
            'Interactively deselect the temperatures to ignore\n'
            'Produces same window as --select_T, but with all temperatures\n'
            'pre-selected.'
        )
    )

    ac_parser.add_argument(
        '--data_header',
        metavar='<str>',
        type=str,
        default='[Data]',
        help=(
            'String used to locate start of data in ac.dat file '
            'Default: [Data]'
        )
    )
    ac_parser.add_argument(
        '--single_plots',
        type=str,
        choices=['on', 'show', 'save', 'off'],
        default='off',
        help=(
            'Plot each experiment and colecole/susceptibility fit separately\n'
            ' - \'on\' shows and saves the plots\n'
            ' - \'show\' shows the plots\n'
            ' - \'save\' saves the plots\n'
            ' - \'off\' neither shows nor saves\n'
            'Default: off (i.e. disabled)'
        )
    )
    # Error defined as
    # sqrt( sum(square( linear(val_x, *linear_popt) -  val_y )) ).
    # The larger this value, the tighter the constraint.\n
    ac_parser.add_argument(
        '--flat_thresh',
        metavar='<Value>',
        type=float,
        default=1E-06,
        help='Threshold to discard flat lines. Default: 1E-06'
    )

    ac_parser.add_argument(
        '--verbose',
        action='store_true',
        help='Print the read-in values from the file indicated'
    )

    # Waveform
    description_Waveform = '''
    Extract AC susceptibilities from waveform data
    See PCCP, 2019, 21, 22302-22307 for method details
    '''
    epilog_Waveform = '''
    This module creates a $NAME_toccfit.dat file to be passed to the AC module
    '''
    waveform_parser = subparsers.add_parser(
        'waveform',
        description=description_Waveform,
        epilog=epilog_Waveform,
        formatter_class=argparse.RawTextHelpFormatter
    )

    waveform_parser.set_defaults(func=waveform_mode_func)

    waveform_parser.add_argument(
        'input_file',
        type=str,
        metavar='<filename(s)>',
        nargs='+',
        help=(
            'SQUID file(s), one per frequency.\n'
            'Supports shell-style wildcards, e.g. data_*.dat'
        )
    )
    waveform_parser.add_argument(
        '--data_header',
        metavar='<str>',
        type=str,
        default='[Data]',
        help='String used to locate data in <filename(s)>. Default: [Data]'
    )
    waveform_parser.add_argument(
        '--field_window',
        metavar='<float>',
        type=float,
        nargs=2,
        default=[-0.5, 0.5],
        help='''Min & max field values (Oe) used to define a waveform block
        Default: -0.5 0.5'''
    )
    waveform_parser.add_argument(
        '--ft_plots',
        choices=['on', 'show', 'save', 'off'],
        metavar='<str>',
        type=str,
        default='off',
        help=(
            'Plot Fourier transform of each individual waveform experiment\n'
            ' - \'on\' shows and saves the plots\n'
            ' - \'show\' shows the plots\n'
            ' - \'save\' saves the plots\n'
            ' - \'off\' neither shows nor saves\n'
            'Default: off (i.e. disabled)'
        )
    )
    waveform_parser.add_argument(
        '--susc_plot',
        choices=['on', 'show', 'save', 'off'],
        metavar='<str>',
        type=str,
        default='on',
        help=(
            'Plot Susceptibility data obtained from the Fourier transform\n'
            ' - \'on\' shows and saves the plots\n'
            ' - \'show\' shows the plots\n'
            ' - \'save\' saves the plots\n'
            ' - \'off\' neither shows nor saves\n'
            'Default: off (i.e. disabled)'
        )
    )
    waveform_parser.add_argument(
        '--temp_thresh',
        metavar='<float>',
        type=float,
        default=0.1,
        help='Threshold used to discriminate between temperatures, default 0.1 K' # noqa
    )
    # DC mode
    description_dc = '''
    Extract relaxation times from magnetisation decays using exponentials and
    (optional) fit the resulting relaxation profile
    '''
    dc_parser = subparsers.add_parser(
        'dc',
        description=description_dc,
        formatter_class=argparse.RawTextHelpFormatter
    )

    dc_parser.set_defaults(func=dc_mode_func)

    dc_parser.add_argument(
        'input_file',
        metavar='<filename>',
        nargs='+',
        type=str,
        help='''File(s) containing the DC magnetisation decay data.
        See Manual for expected format.'''
    )
    dc_parser.add_argument(
        '--temp_thresh',
        metavar='<float>',
        type=float,
        default=0.1,
        help='Threshold used to discriminate between temperatures, default 0.1 K' # noqa
    )
    dc_parser.add_argument(
        '--process',
        metavar='<Option>',
        choices=['plot', 'decays', 'all'],
        default='all',
        help=(
            'What to do:\n'
            ' - \'plot\' Show only raw magnetisation decays\n'
            ' - \'decays\' - Fit only the magnetisation decays\n'
            ' - \'all\' - Fit magnetisation decays and relaxation profile\n'
            'Default: all\n'
        )
    )
    dc_parser.add_argument(
        '--model',
        metavar='<Option>',
        choices=['stretched', 'double'],
        default='stretched',
        help=(
            'Select which model to apply to each decay trace\n'
            ' - \'stretched\' - Fit to a single stretched exponential\n'
            ' - \'double\' - Fit to a double stretched exponential\n'
            'Default: stretched.\n'
        )
    )

    meq_values = dc_parser.add_mutually_exclusive_group()

    meq_values.add_argument(
        '--M_eq',
        metavar='<fit/fix VALUE>',
        default=['fix', 0.],
        action=FitFixAction,
        nargs=2,
        help=(
            'Controls fitting/fixing of M_eq.\n'
            'First value is either the word fit or fix\n'
            'Second value is either the initial value  or fixed value as :\n'
            ' - The word guess - uses the final measured moment for each experiment (after cutting with --cut_moment)\n' # noqa
            ' - A single number specifying a single value to use for all experiments\n' # noqa
            ' - A filename, where the file contains a column of parameter values, one per experiment \n' # noqa
            'Default: fix 0.\n'
        )
    )

    meq_values.add_argument(
        '--M_eq_calibration',
        metavar='<Value>',
        action=MultiAction,
        nargs=3,
        default=None,
        help=(
            'Calibrate M_eq for infield decays where sample mass or Molecular mass is unknown\n' # noqa
            'First value is the filename for the experimental saturation magnetisations\n'  # noqa
            'Second value is the filename for the theoretical magnetisation curves\n'  # noqa
            'Third value is the saturation magnetisation in Oe for all experiments\n' # noqa
            'Default: None'
        )
    )

    dc_parser.add_argument(
        '--M_0',
        metavar='<fit/fix VALUE>',
        default=['fix', 'guess'],
        action=FitFixAction,
        nargs=2,
        help=(
            'Controls fitting/fixing of M_0.\n'
            'First value is either the word fit or fix\n'
            'Second value is either the initial value  or fixed value as :\n'
            ' - The word guess - uses the first measured moment for each experiment\n' # noqa
            ' - A single number specifying a single value to use for all experiments\n' # noqa
            ' - A filename, where the file contains a column of parameter values, one per experiment \n' # noqa
            'Default: fix guess\n'
        )
    )

    dc_parser.add_argument(
        '--t_offset',
        metavar='<fit/fix VALUE>',
        default=['fix', 0.],
        action=FitFixAction,
        nargs=2,
        help=(
            'Controls fitting/fixing of t_offset.\n'
            'First value is either the word fit or fix\n'
            'Second value is either the initial value  or fixed value as :\n'
            ' - A single number specifying a single value to use for all experiments\n' # noqa
            ' - A filename, where the file contains a column of parameter values, one per experiment \n' # noqa
            'Default: fix 0.\n'
        )
    )

    dc_parser.add_argument(
        '--taustar',
        metavar='<fit/fix VALUE>',
        default=['fit', 100.],
        action=FitFixAction,
        nargs=2,
        help=(
            'Controls fitting/fixing of tau* in Stretched Exponential model.\n'
            'First value is either the word fit or fix\n'
            'Second value specifies the initial value  or fixed value as :\n'
            ' - A single number specifying a single value to use for all experiments\n' # noqa
            ' - A filename, where the file contains a column of parameter values, one per experiment \n' # noqa
            'Default: fit 100.\n'
        )
    )

    dc_parser.add_argument(
        '--beta',
        metavar='<fit/fix VALUE>',
        default=['fit', 0.95],
        action=FitFixAction,
        nargs=2,
        help=(
            'Controls fitting/fixing of beta in Stretched Exponential model.\n'
            'First value is either the word fit or fix\n'
            'Second value is either the initial value  or fixed value as :\n'
            ' - A single number specifying a single value to use for all experiments\n' # noqa
            ' - A filename, where the file contains a column of parameter values, one per experiment \n' # noqa
            'Default: fit 0.95\n'
        )
    )

    dc_parser.add_argument(
        '--taustar1',
        metavar='<fit/fix VALUE>',
        default=['fit', 50.],
        action=FitFixAction,
        nargs=2,
        help=(
            'Controls fitting/fixing of tau_1* in Double Exponential Model.\n'
            'First value is either the word fit or fix\n'
            'Second value specifies the initial value  or fixed value as :\n'
            ' - A single number specifying a single value to use for all experiments\n' # noqa
            ' - A filename, where the file contains a column of parameter values, one per experiment \n' # noqa
            'Default: fit 50\n'
        )
    )

    dc_parser.add_argument(
        '--taustar2',
        metavar='<fit/fix VALUE>',
        default=['fit', 5000.],
        action=FitFixAction,
        nargs=2,
        help=(
            'Controls fitting/fixing of tau_2* in Double Exponential Model.\n'
            'First value is either the word fit or fix\n'
            'Second value specifies the initial value  or fixed value as :\n'
            ' - A single number specifying a single value to use for all experiments\n' # noqa
            ' - A filename, where the file contains a column of parameter values, one per experiment \n' # noqa
            'Default: fit 5000.\n'
        )
    )

    dc_parser.add_argument(
        '--beta1',
        metavar='<fit/fix VALUE>',
        default=['fit', 0.95],
        action=FitFixAction,
        nargs=2,
        help=(
            'Controls fitting/fixing of beta_1 in Double Exponential Model.\n'
            'First value is either the word fit or fix\n'
            'Second value is either the initial value  or fixed value as :\n'
            ' - A single number specifying a single value to use for all experiments\n' # noqa
            ' - A filename, where the file contains a column of parameter values, one per experiment \n' # noqa
            'Default: fit 0.95\n'
        )
    )

    dc_parser.add_argument(
        '--beta2',
        metavar='<fit/fix VALUE>',
        default=['fit', 0.95],
        action=FitFixAction,
        nargs=2,
        help=(
            'Controls fitting/fixing of beta_2 in Double Exponential Model.\n'
            'First value is either the word fit or fix\n'
            'Second value is either the initial value  or fixed value as :\n'
            ' - A single number specifying a single value to use for all experiments\n' # noqa
            ' - A filename, where the file contains a column of parameter values, one per experiment \n' # noqa
            'Default: fit 0.95\n'
        )
    )

    dc_parser.add_argument(
        '--frac',
        metavar='<fit/fix VALUE>',
        default=['fit', 0.5],
        action=FitFixAction,
        nargs=2,
        help=(
            'Controls fitting/fixing of fraction in Double Exponential Model.\n' # noqa
            'First value is either the word fit or fix\n'
            'Second value is either the initial value or fixed value as :\n'
            ' - A single number specifying a single value to use for all experiments\n' # noqa
            ' - A filename, where the file contains a column of parameter values, one per experiment \n' # noqa
            'Default: fit 0.5\n'
        )
    )

    dc_parser.add_argument(
        '--cut_moment',
        metavar='<number>',
        type=float,
        default=0.01,
        help=(
            'Number specifying a %% of initial moment.\n'
            'Moments smaller than this will be discarded.\n'
            'Disable by setting to 0.\n'
            'Default: 0.01'
        )
    )
    dc_parser.add_argument(
        '--field_calibration',
        metavar='<VALUE>',
        action=DiffAction,
        default=None,
        help=(
            'Set what the target field has been calibrated to. Options are:\n'
            ' - A single value giving the actual field for all experiments.\n'
            ' - A filename, where the file contains a column of parameter values, one per experiment \n' # noqa
            'Default: None.\n'
        )
    )

    field_cutting = dc_parser.add_mutually_exclusive_group()

    field_cutting.add_argument(
        '--dfield_thresh',
        metavar='<number>',
        type=float,
        default=0.5,
        help=(
            'Threshold (in Oe) used for finding stable field\n'
            'Defined as difference between DC fields, below which field is '
            'considered stable\n'
            'Default: 0.5 Oe'
        )
    )

    field_cutting.add_argument(
        '--no_field_discard',
        action='store_true',
        help='''Do not cut data at stable field'''
    )

    dc_parser.add_argument(
        '--data_header',
        metavar='<str>',
        type=str,
        default='[Data]',
        help=(
            'String used to locate start of data in dc.dat file '
            'Default: [Data]'
        )
    )

    dc_parser.add_argument(
        '--moment_header',
        metavar='<str>',
        type=str,
        choices=['find', 'moment', 'fixed', 'free'],
        help=(
            'Controls which column of datafile is used for magnetic moment.\n'
            'Choices are:\n'
            ' - find - ccfit2 will use the first non-empty column it can find\n' # noqa
            ' - moment - ccfit2 will use the \'Moment (emu) column\'\n'
            ' - fixed - ccfit2 will use the \'DC Moment Fixed Ctr (emu)\' column\n' # noqa
            ' - free - ccfit2 will use the \'DC Moment Free Ctr (emu)\' column\n' # noqa
            'Default: find\n'
        )
    )

    dc_parser.add_argument(
        '--decay_plots',
        type=str,
        choices=['on', 'show', 'save', 'off'],
        default='show',
        help=(
            'Visualisation options for individual magnetisation decay plots\n'
            ' - \'on\' shows and saves the plots\n'
            ' - \'show\' shows the plots\n'
            ' - \'save\' saves the plots\n'
            ' - \'off\' neither shows nor saves\n'
            'Default: show (i.e. disabled)'
        )
    )

    dc_parser.add_argument(
        '--hide_params',
        action='store_false',
        help='''Hides parameters on magnetisation decay plots.'''
    )

    dc_parser.add_argument(
        '--plot_axes',
        metavar='<str>',
        type=str,
        nargs=2,
        default=['linear', 'linear'],
        choices=['linear', 'log'],
        help=(
            'Choose between linear and log axes for plotting the DC decay '
            'traces.\n'
            'Default: linear linear.'
        )
    )
    dc_parser.add_argument(
        '--verbose',
        action='store_true',
        help='Print guess information to screen'
    )

    # Relaxation Profile
    description_relaxation = '''
    Fit the field- and/or temperature-dependence of relaxation times.
    '''
    relaxation_parser = subparsers.add_parser(
        'relaxation',
        description=description_relaxation,
        formatter_class=argparse.RawTextHelpFormatter
    )

    relaxation_parser.set_defaults(func=relaxation_mode_func)

    relaxation_parser.add_argument(
        'input_files',
        type=str,
        nargs='+',
        help=(
            'Either AC or DC <>_params.out files.'
            'See documentation for format'
        )
    )
    relaxation_parser.add_argument(
        '--x_var',
        metavar='<Option>',
        choices=['T', 'H'],
        default='T',
        help=(
            'Fit either the temperature (T) or field (H) dependence of '
            'rates (isofield or isothermal).\n'
            'Options: T, H\n'
            'Default: T'
        )
    )
    relaxation_parser.add_argument(
        '--process',
        metavar='<Option>',
        choices=['plot', 'all'],
        default='all',
        help=(
            'What to do:\n'
            ' - \'plot\' Show only raw relaxation profile\n'
            ' - \'all\' - Fit relaxation profile\n'
            'Default: all\n'
        )
    )

    relaxation_parser.add_argument(
        '--filetype',
        metavar='<Option>',
        choices=['ccfit2', 'rate'],
        default='ccfit2',
        help=(
            'Type of file to read. Options:\n'
            ' - ccfit2 *_params.out file\n'
            ' - rates file (see docs)\n'
            'Default: ccfit2'
        )
    )

    relaxation_parser.add_argument(
        '--no_weights',
        action='store_true',
        help='Disable weighting of fit using uncertainties'
    )

    relaxation_parser.add_argument(
        '--verbose',
        action='store_true',
        help='Print the read-in values from the file indicated.'
    )

    relaxation_parser.add_argument(
        '--plot_type',
        metavar='<Option>',
        choices=['rate', 'time', 'both'],
        default='rate',
        help=(
            'Type of final relaxation plot to produce\n'
            ' - \'rate\' rate vs T or H on a log-log or log-symlog plot\n'
            ' - \'time\' ln(time) vs 1/T or 1/H on a lin-lin plot\n'
            ' - \'both\' produce both of the above\n'
            'Default: rate'
        )
    )

    relaxation_parser.add_argument(
        '--hide_params',
        action='store_false',
        help='''Hides parameters on fitted relaxation plots.'''
    )

    # If argument list is none, then call function func
    # which is assigned to help function
    # read sub-parser
    parser.set_defaults(func=lambda args: parser.print_help())
    known_args = parser.parse_args(arg_list)

    known_args.func(known_args)

    return known_args


def main():
    read_args()



from scipy.linalg import svd
from scipy.optimize._optimize import OptimizeResult
import numpy as np


def svd_stdev(curr_fit: OptimizeResult) -> tuple[list[float], list[bool]]:
    '''
    Calculates standard deviation of fit-parameters given output of scipy
    least_squares.

    Uses SVD of jacobian to check for singlular values equal to zero.
    Singular values equal to zero are discarded, so the corresponding
    parameter has a standard deviation which cannot be computed and is
    instead set to zero
    (i.e. they are meaningless)

    Parameters
    ----------
    curr_fit: OptimizeResult
        Result object from scipy.optimise.least_squares

    Returns
    -------
    list[float]
        Standard deviation on parameters
    list[bool]
        bool for each parameter, if False, then standard deviation cannot be
        calculated (and is set to zero)
    '''

    # SVD of jacobian
    _, s, VT = svd(curr_fit.jac, full_matrices=False)
    # Zero threshold as multiple of machine precision
    threshold = np.finfo(float).eps * max(curr_fit.jac.shape) * s[0]
    # Find singular values = 0.
    nonzero_sing = s > threshold
    # Truncate to remove these values
    s = s[nonzero_sing]
    VT = VT[:s.size]
    # Calculate covariance of each parameter using truncated arrays
    pcov = VT.T / s**2 @ VT
    # Scale by reduced chi**2 to remove influence of input sigma (if present)
    # and just obtain standard deviation of fit
    chi2dof = np.sum(curr_fit.fun**2)
    chi2dof /= (curr_fit.fun.size - curr_fit.x.size)
    pcov *= chi2dof
    stdev = np.sqrt(np.diag(pcov))

    no_stdev = stdev > threshold

    if sum(nonzero_sing) == len(nonzero_sing):
        no_stdev = [True] * len(nonzero_sing)

    return stdev, no_stdev



'''
This module contains functions and objects for working with DC
magnetisation decay data
'''

from math import isnan
import numpy as np
import numpy.typing as npt
import matplotlib.pyplot as plt
from matplotlib.ticker import AutoMinorLocator
from scipy.optimize import least_squares
from abc import ABC, abstractmethod
import copy
import operator
import warnings

from . import utils as ut
from . import stats

#: Supported DC Headers - One of each MUST be found in the input file.
#:
#:  Note - These differ between magnetometers\
#:
#:  These keys are the arguments to the Measurement constructor, but their
#:  order does not matter
HEADERS_SUPPORTED: dict[str, list[str]] = {
    'dc_field': [
        'Magnetic Field (Oe)',
        'Field (Oe)'
    ],
    'temperature': [
        'Temperature (K)'
    ],
    'time': [
        'Time Stamp (sec)'
    ],
    'moment': [
        'Moment (emu)',
        'DC Moment Free Ctr (emu)',
        'DC Moment Fixed Ctr (emu)'
    ]
}

# Generic dc magnetometer file header names
HEADERS_GENERIC = HEADERS_SUPPORTED.keys()


class Measurement():
    '''
    Stores data for a single DC Decay measurement at a
    given temperature and applied field

    Parameters
    ----------
    dc_field : float
        Applied dc field (Oe)
    temperature : float
        Temperature of datapoint (K)
    moment : float
        Magnetic moment of datapoint (emu)
    time : float
        Time of datapoint (s)

    Attributes
    ----------
    dc_field : float
        Applied dc field (Oe)
    temperature : float
        Temperature of datapoint (K)
    moment : float
        Magnetic moment of datapoint (emu)
    time : float
        Time of datapoint (s)
    rep_temperature : float
        Representative temperature assigned to this datapoint (K)
    rep_dc_field : float
        Representative dc field assigned to this datapoint (Oe)
    '''

    def __init__(self, dc_field: float, temperature: float, time: float,
                 moment: float):
        self.dc_field = dc_field
        self.temperature = temperature
        self.time = time
        self.moment = moment
        self._rep_temperature = None
        self._rep_dc_field = None

    @property
    def rep_temperature(self):
        return self._rep_temperature

    @rep_temperature.setter
    def rep_temperature(self, value: float):
        self._rep_temperature = value
        return

    @property
    def rep_dc_field(self):
        return self._rep_dc_field

    @rep_dc_field.setter
    def rep_dc_field(self, value: float):
        self._rep_dc_field = value
        return

    @classmethod
    def from_file(cls, file: str, header_indices: dict | str = 'find',
                  data_header: str = '[Data]',
                  verbose: bool = True) -> list['Measurement']:
        '''
        Extracts dc data from magnetometer output file and
        returns list of datapoints, one for each valid measurement.\n
        Incomplete lines are ignored

        Parameters
        ----------
        file : str
            Name of magnetometer output file
        header_indices : str | dict, default 'find'
            Default 'find' will automatically locate headers, else provide dict
            with:\n
            Keys as generic header names given in `HEADERS_GENERIC`\n
            Values as column index (number) of header in file\n
        data_header : str default '[Data]'
            Contents of line which specifies the beginning of the data block
            in input file.\n
            Default is to find line containing '[Data]'
        verbose: bool, default True
            If True, issues parsing measurements are written to terminal
        Returns
        -------
        list
            Measurement objects, one per temperature, per field\n
            List has the same order as the magnetometer file

        '''

        encoding = ut.detect_encoding(file)

        # Check data_header is in file
        data_index = ut.locate_data_header(file, data_header=data_header)

        if header_indices == 'find':
            # Get file headers
            header_indices, _ = ut.parse_headers(
                file, data_index, HEADERS_SUPPORTED
            )

        # Columns to extract from file
        cols = {
            gen: header_indices[gen] for gen in HEADERS_GENERIC
        }

        # Convert strings to floats, if not possible then mark as NaN
        converters = {
            it: lambda s: (float(s.strip() or np.NaN)) for it in cols.values()
        }

        # Read required columns of file
        data = np.loadtxt(
            file,
            skiprows=data_index + 1,
            delimiter=',',
            converters=converters,
            usecols=cols.values(),
            encoding=encoding
        )

        # Remove missing entries that have been marked as nan
        data = [
            row for row in data
            if not any(isnan(val) for val in row)
        ]

        # Convert array of floats into list of Measurement objects, one per
        # line
        # Remove positional nature of Measurement constructor args by using
        # kwargs through dict
        measurements = [
            cls(**{
                col: val
                for col, val in zip(cols, row)
            })
            for row in data
            if not any(isnan(val) for val in row)
        ]

        if not len(measurements) and verbose:
            _msg = '\n Error: Cannot parse measurements from file {}'.format(
                file
            )
            ut.cprint(_msg, 'red')

        return measurements


class Experiment():
    '''
    Stores data for multiple DC measurements at a
    given temperature, given applied dc field

    Parameters
    ----------
    rep_temperature: float
        Representative temperature of experiment (K)
        e.g. mean of all datapoints (Measurements)
    rep_dc_field : float
        Representative dc field assigned to this experiment (Oe)
        e.g. mean of all datapoints
    raw_temperatures: array_like
        Raw temperatures of experiment, one per datapoint (K)
    times : array_like
        Time value, one per datapoint (s)
    moments : array_like
        Measured moment, one value per datapoint (emu)
    dc_fields : array_like
        Applied dc field in Oe, one value per datapoint (Oe)

    Attributes
    ----------
    rep_temperature: float
        Representative temperature of experiment
        e.g. mean of all datapoints (Measurements)
    rep_dc_field : float
        Representative dc field assigned to this experiment (Oe)
        e.g. mean of all datapoints
    raw_temperatures: ndarray of floats
        Raw temperatures of experiment, one per datapoint (K)
    times : ndarray of floats
        Time value, one per datapoint (s)
    moments : ndarray of floats
        Measured moment, one value per datapoint (emu)
    dc_fields : ndarray of floats
        Applied dc field in Oe, one value per datapoint (Oe)
    meas_dc_fields : ndarray of floats
        Applied dc field measured by the instrument\n
        one value per datapoint (Oe)
    '''

    def __init__(self, rep_temperature: float, rep_dc_field: float,
                 raw_temperatures: npt.ArrayLike,
                 times: npt.ArrayLike, moments: npt.ArrayLike,
                 dc_fields: npt.ArrayLike):

        self.rep_temperature = rep_temperature
        self.rep_dc_field = rep_dc_field
        self.raw_temperatures = np.asarray(raw_temperatures)
        self.times = np.asarray(times)
        self.moments = np.asarray(moments)
        self.dc_fields = np.asarray(dc_fields)
        self._meas_dc_fields = np.copy(self.dc_fields)

        return

    @property
    def meas_dc_fields(self) -> npt.NDArray:
        return self._meas_dc_fields

    @meas_dc_fields.setter
    def meas_dc_fields(self, value: npt.ArrayLike):
        self._meas_dc_fields = np.asarray(value)
        return

    @classmethod
    def _from_single(cls, measurements: list[Measurement],
                     cut_moment: float = 0.01) -> 'Experiment':
        '''
        Creates experiment from single set of measurements

        Parameters
        ----------
        measurements: list[Measurement]
            Measurements constituting a single experiment
        cut_moment: float, default 0.01
            Specifies %% of initial moment\n
            Moments smaller than this will be discarded
        Returns
        -------
            Experiment
        '''

        # Remove moments less than cut_moment % of initial moment
        measurements = [
            mm
            for mm in measurements
            if mm.moment >= cut_moment * measurements[0].moment
        ]

        raw_temperatures = np.array([
            mm.temperature for mm in measurements
        ])

        moments = np.array([
            mm.moment for mm in measurements
        ])

        # Set start time to zero
        mintime = np.min([
            mm.time
            for mm in measurements
        ])

        # Set start time of these measurements as zero
        for mm in measurements:
            mm.time -= mintime

        times = np.array([
            mm.time for mm in measurements
        ])

        dc_fields = np.array([
            mm.dc_field for mm in measurements
        ])

        rep_temperature = measurements[0].rep_temperature
        rep_dc_field = measurements[0].rep_dc_field

        experiment = cls(
            rep_temperature,
            rep_dc_field,
            raw_temperatures,
            times,
            moments,
            dc_fields
        )

        return experiment

    @classmethod
    def from_measurements(cls, measurements: list[Measurement],
                          temp_thresh: float = 0.1,
                          field_thresh: float = 1,
                          cut_moment: float = 0.01) -> list['Experiment']:
        '''
        Creates list of Experiment objects from a list of individual
        Measurement objects. Experiments are defined as a set of measurements
        with the same mean temperature and mean field - defined by thresholds

        Parameters
        ----------
        measurement: list[Measurement]
            Measurements at various temperatures and times
        temp_thresh: float, default 0.1 K
            Threshold used to discriminate between temperatures (K)
        field_thresh: float, default 1 Oe
            Threshold used to discriminate between dc field values (Oe)
        cut_moment: float, default 0.01
            Specifies %% of initial moment\n
            Moments smaller than this will be discarded
            after splitting by field and temperature.

        Returns
        -------
        list[Experiment]
            Each element is an Experiment\n
            The list is sorted by\n
            1. Mean dc_field (low to high)\n
            2. then by mean temperature (low to high)\n
        '''

        # Sort measurements by dc field, temperature, and time
        measurements = sorted(
            measurements,
            key=operator.attrgetter('dc_field', 'temperature', 'time')
        )

        # Find mean field values
        mean_fields, fsi = ut.find_mean_values(
            [
                measurement.dc_field
                for measurement in measurements
            ],
            thresh=field_thresh
        )

        # Set each measurement's representative dc field, here the mean
        for measurement, mean_field in zip(measurements, np.concatenate(mean_fields)): # noqa
            measurement.rep_dc_field = mean_field

        # Re-sort using mean field
        measurements = sorted(
            measurements,
            key=operator.attrgetter('rep_dc_field', 'temperature', 'time')
        )
        # Get indices which sort the above
        order = sorted(
            np.arange(len(measurements), dtype=int),
            key=lambda k: (
                measurements[k].rep_dc_field,
                measurements[k].temperature,
                measurements[k].time
            )
        )
        # transfer field split points into new order
        fsi = np.array([order[fs] for fs in fsi], dtype=int)

        # Find mean temperature values
        mean_temperatures, tsi = ut.find_mean_values(
            [
                measurement.temperature
                for measurement in measurements
            ],
            thresh=temp_thresh
        )

        # Set each measurement's representative temperature, here the mean
        for m, mt in zip(measurements, np.concatenate(mean_temperatures)):
            m.rep_temperature = mt

        # Re-sort using mean dc field and mean temperature
        measurements = sorted(
            measurements,
            key=operator.attrgetter('rep_dc_field', 'rep_temperature', 'time')
        )
        # Get indices which sort the above
        order = sorted(
            np.arange(len(measurements), dtype=int),
            key=lambda k: (
                measurements[k].rep_dc_field,
                measurements[k].rep_temperature,
                measurements[k].time
            )
        )
        # transfer temperature split points into new order
        tsi = np.array([order[ts] for ts in tsi], dtype=int)
        # transfer field split points into new order
        fsi = np.array([order[fs] for fs in fsi], dtype=int)

        # Split and field and temperature jumps
        split_ind = np.concatenate([tsi, fsi])
        # Remove duplicate split indices corresponding to field and temperature
        # jump
        split_ind = np.unique(split_ind)

        # If only one experiment, then no need to split, just
        # combine measurements into an experiment
        if not len(split_ind):
            experiments = [
                cls._from_single(measurements, cut_moment=cut_moment)
            ]
        # >1 Experiment, then split up measurements from time jumps
        else:
            split_measurements: list[list[Measurement]] = np.split(
                measurements,
                split_ind + 1
            )

            # and combine each set of measurements into an experiment
            experiments = [
                cls._from_single(sm, cut_moment=cut_moment)
                for sm in split_measurements
            ]

        return experiments


class Model(ABC):
    '''
    Abstract class on which all models of DC magnetisation decays are based
    '''

    @property
    @abstractmethod
    def NAME() -> str:
        'String name of model'
        raise NotImplementedError

    @property
    @abstractmethod
    def DISP_NAME() -> str:
        'Display name for interactive buttons'
        raise NotImplementedError

    @property
    @abstractmethod
    def PARNAMES() -> list[str]:
        'String names of parameters which can be fitted or fixed'
        raise NotImplementedError

    @property
    @abstractmethod
    def VARNAMES_MM() -> dict[str, str]:
        '''
        Mathmode (i.e. $$, latex ) versions of PARNAMES\n
        Keys are strings from PARNAMES plus any other variables which
        might be plotted (e.g. lntau_expect)\n
        Values are mathmode strings
        '''
        raise NotImplementedError

    @property
    @abstractmethod
    def UNITS() -> dict[str, str]:
        '''
        string names of units of PARNAMES\n
        Keys are strings from PARNAMES
        might be needed\n
        Values are unit name strings
        '''
        raise NotImplementedError

    @property
    @abstractmethod
    def UNITS_MM() -> dict[str, str]:
        '''
        Mathmode (i.e. $$, latex ) versions of UNITS\n
        Keys are strings from PARNAMES
        might be needed\n
        Values are unit name strings
        '''
        raise NotImplementedError

    @property
    @abstractmethod
    def BOUNDS() -> dict[str, list[float, float]]:
        '''
        Bounds for each parameter of model
        keys: parameter name
        values: [upper, lower]
        used by scipy least_squares
        '''
        raise NotImplementedError

    @abstractmethod
    def calc_lntau_expect() -> float | list[float]:
        '''
        Calculates expectation value of ln(tau) from input parameters
        '''
        raise NotImplementedError

    @abstractmethod
    def calc_lntau_fit_ul() -> list[float]:
        '''
        Calculates upper and lower bounds of ln(tau) from fit uncertainty
        in fitted parameters
        '''
        raise NotImplementedError

    @abstractmethod
    def calc_lntau_stdev() -> float | list[float]:
        '''
        Calculates standard deviation of ln(tau) from input paramaters
        '''
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def model(parameters: dict[str, float], time: list[float]) -> list[float]:
        '''
        Evaluates model function of DC magnetisation decay
        using provided parameter and time values.

        Parameters
        ----------
        parameters: list[float]
            Parameters to use in model function
        time: list[float]
            Time values in seconds at which model function is evaluated

        Returns
        -------
        list[float]
            Moment as a function of time

        '''
        raise NotImplementedError

    @classmethod
    def residuals(cls, params: dict[str, float], time: list[float],
                  true_moment: list[float]) -> list[float]:
        '''
        Calculates difference between measured moment and
        trial moment from model

        Parameters
        ----------
        params : list[float]
            model parameter values
        time: list[float]
            time values in seconds at which model function is evaluated
        moment : list[float]
            true (experimental) values of magnetic moment

        Returns
        -------
        list[float]
            Residuals
        '''
        trial_moment = cls.model(params, time)

        residuals = trial_moment - true_moment
        return residuals

    def __init__(self):
        '''
        Set default values for mandatory attributes
        '''

        # Final model parameter values
        self._final_var_values = {
            var: None
            for var in self.PARNAMES
        }

        # Fit status, temperature, and field
        self._fit_status = False
        self._temperature = None
        self._dc_field = None
        self._meas_dc_field = None

        # Fit standard deviation
        self._fit_stdev = {
            var: None
            for var in self.fit_vars.keys()
        }

        # Expectation value and standard deviation of ln(tau)
        self._lntau_expect = None
        self._lntau_fit_ul = None
        self._lntau_stdev = None

        return

    @property
    def fit_status(self) -> bool:
        'True if fit successful, else False'
        return self._fit_status

    @fit_status.setter
    def fit_status(self, value):
        if isinstance(value, bool):
            self._fit_status = value
        else:
            raise TypeError
        return

    @property
    def temperature(self) -> float:
        'Temperature of fit (K)'
        return self._temperature

    @temperature.setter
    def temperature(self, value):
        if isinstance(value, (np.floating, float)):
            self._temperature = value
        else:
            raise TypeError
        return

    @property
    def dc_field(self) -> float:
        'DC Field of fit (Oe)'
        return self._dc_field

    @dc_field.setter
    def dc_field(self, value):
        if isinstance(value, (np.floating, float)):
            self._dc_field = value
        else:
            raise TypeError
        return

    @property
    def meas_dc_field(self) -> float:
        'Measured DC Field of fit (Oe)'
        return self._meas_dc_field

    @meas_dc_field.setter
    def meas_dc_field(self, value: float):
        if isinstance(value, (np.floating, float)):
            self._meas_dc_field = value
        else:
            raise TypeError
        return

    @property
    def final_var_values(self) -> float:
        'Final values of parameters, both fitted and fixed'
        return self._final_var_values

    @final_var_values.setter
    def final_var_values(self, value):
        if isinstance(value, dict):
            self._final_var_values = value
        else:
            raise TypeError
        return

    @property
    def fit_stdev(self) -> float:
        'Standard deviation on fitted parameters, from fitting routine'
        return self._fit_stdev

    @fit_stdev.setter
    def fit_stdev(self, value):
        if isinstance(value, dict):
            self._fit_stdev = value
        else:
            raise TypeError
        return

    @property
    def fix_vars(self) -> dict[str, float]:
        '''
        Parameters to fix (i.e. not fit)
        keys are names in PARNAMES, values are values
        '''
        return self._fix_vars

    @fix_vars.setter
    def fix_vars(self, value):
        if isinstance(value, dict):
            if any([key not in self.PARNAMES for key in value.keys()]):
                raise KeyError('Unknown variable names provided to fix')
            self._fix_vars = value
        else:
            raise TypeError('fix must be dictionary')
        return

    @property
    def fit_vars(self) -> dict[str, float]:
        '''
        Parameters to fit
        keys are names in PARNAMES, values are values
        '''
        return self._fit_vars

    @fit_vars.setter
    def fit_vars(self, value):
        if isinstance(value, dict):
            if any([key not in self.PARNAMES for key in value.keys()]):
                raise KeyError('Unknown variable names provided to fit')
            self._fit_vars = value
        else:
            raise TypeError('Fit must be dictionary')
        return

    @property
    def lntau_stdev(self) -> float:
        'Standard deviation of ln(tau)'
        # If not calculated yet, then calculate
        if self._lntau_stdev is None:
            self.lntau_stdev = self.calc_lntau_stdev()
        return self._lntau_stdev

    @lntau_stdev.setter
    def lntau_stdev(self, value):
        self._lntau_stdev = value

    @property
    def lntau_fit_ul(self) -> list[float]:
        '''
        Upper and lower (1 sigma) limits of ln(tau) from fit uncertainty
        in fitted values
        '''
        # If not calculated yet, then calculate
        if self._lntau_fit_ul is None:
            self.lntau_fit_ul = self.calc_lntau_fit_ul()
        return self._lntau_fit_ul

    @lntau_fit_ul.setter
    def lntau_fit_ul(self, value):
        self._lntau_fit_ul = value

    @property
    def lntau_expect(self) -> float:
        'Expectation value of ln(tau)'
        # If not calculated yet, then calculate
        if self._lntau_expect is None:
            self.lntau_expect = self.calc_lntau_expect()
        return self._lntau_expect

    @lntau_expect.setter
    def lntau_expect(self, value):
        self._lntau_expect = value

    @classmethod
    def residual_from_float_list(cls, new_vals: list[float],
                                 fit_vars: dict[str, float],
                                 fix_vars: dict[str, float],
                                 time: list[float],
                                 moment: list[float]) -> list[float]:
        '''
        Wrapper for `residuals` method, takes new values from fitting routine
        which provides list[float], to construct new fit_vals dict, then
        runs `residuals` method.

        Parameters
        ----------

        fit_vars: dict[str, float]
            Parameter to fit in model function\n
            keys are PARNAMES, values are float values
        fix_vars: dict[str, float]
            Parameter which remain fixed in model function\n
            keys are PARNAMES, values are float values
        time: list[float]
            time values in seconds at which model function is evaluated
        moment : list[float]
            true (experimental) values of magnetic moment

        Returns
        -------
        list[float]
            Residuals
        '''

        # Swap fit values for new values from fit routine
        new_fit_vars = {
            name: guess
            for guess, name in zip(new_vals, fit_vars.keys())
        }

        # And make combined dict of fit and fixed
        # variable names (keys) and values
        all_vars = {**fix_vars, **new_fit_vars}

        return cls.residuals(all_vars, time, moment)

    def fit_to(self, experiment: 'Experiment',
               guess: str | list[float] = 'predefined',
               verbose: bool = True) -> None:
        '''
        Fits model to Experiment

        Parameters
        ----------
        experiment : Experiment
            Experiment to which a model will be fitted
        guess : str | list[float], default 'predefined'
            Either string 'predefined', or dict of initial parameters
            used as starting guess with keys as parameter names and values as
            numerical values
        verbose: bool, default True
            If True, prints information to terminal

        Returns
        -------
        None
        '''

        # Get starting guesses, either predefined in self or given
        # by user
        if guess != 'predefined':
            guess = [
                guess[key] for key in self.fit_vars.keys()
            ]
        else:
            guess = [val for val in self.fit_vars.values()]

        # Get bounds for variables to be fitted
        bounds = np.array([
            self.BOUNDS[name]
            for name in self.fit_vars.keys()
        ]).T

        curr_fit = least_squares(
            self.residual_from_float_list,
            args=[
                self.fit_vars,
                self.fix_vars,
                experiment.times,
                experiment.moments
            ],
            x0=guess,
            bounds=bounds,
            max_nfev=200 * (len(guess) + 1),
            jac='3-point'
        )

        self.temperature = experiment.rep_temperature
        self.dc_field = experiment.dc_fields[-1]
        self.meas_dc_field = experiment.meas_dc_fields[-1]

        if curr_fit.status == 0:
            if verbose:
                ut.cprint(
                    '\n Fit at {} K and {} Oe failed - Too many iterations'.format( # noqa
                        self.temperature, self.dc_field
                    ),
                    'black_yellowbg'
                )

            # Set fitted values
            self.final_var_values = {
                name: value
                for name, value in zip(self.fit_vars, curr_fit.x)
            }
            # and fixed values
            for key, val in self.fix_vars.items():
                self.final_var_values[key] = val

            self.fit_stdev = {
                label: np.nan
                for label in self.fit_vars.keys()
            }
            self.fit_status = False
        else:
            stdev, nonzero_sing = stats.svd_stdev(curr_fit)

            # Standard deviation error on the parameters
            self.fit_stdev = {
                label: val
                for label, val in zip(self.fit_vars.keys(), stdev)
            }

            self.fit_status = True

            # Report singular values=0 of Jacobian
            # and indicate that std_dev cannot be calculated
            for par, si in zip(self.fit_vars.keys(), nonzero_sing):
                if verbose and not si:
                    ut.cprint(
                        f'Warning: Jacobian is degenerate for {par}, standard deviation cannot be found, and is set to zero\n', # noqa
                        'black_yellowbg'
                    )

            # Set fitted values
            self.final_var_values = {
                name: value
                for name, value in zip(self.fit_vars, curr_fit.x)
            }
            # and fixed values
            for key, val in self.fix_vars.items():
                self.final_var_values[key] = val

        return


class ExponentialModel(Model):
    '''
    Exponential Model of Magnetisation decay with stretching parameters
    '''

    #: Model name
    NAME = 'Exponential'

    #: Display name for interactive buttons
    DISP_NAME = copy.copy(NAME)

    #: Model Parameter name strings
    PARNAMES = [
        'tau*', 'beta', 'm_eq', 'm_0', 't_offset'
    ]

    #: Model Parameter bounds
    BOUNDS = {
        'tau*': [0., np.inf],
        'beta': [0., 1.],
        'm_eq': [-np.inf, np.inf],
        'm_0': [0, np.inf],
        't_offset': [0., np.inf]
    }

    #: Model Parameter mathmode name strings
    VARNAMES_MM = {
        'tau*': r'$\tau^*$',
        'lntau_expect': r'$\langle \ln \tau \rangle$',
        'lntau_stdev': r'$\sigma_{\ln \tau}$',
        'beta': r'$\beta$',
        'm_eq': r'$M_\mathregular{eq}$',
        'm_0': r'$M_0$',
        't_offset': r'$\mathregular{t}_\mathregular{offset}$'
    }

    #: Model Parameter Unit strings
    UNITS = {
        'tau*': 's',
        'beta': '',
        'm_eq': 'emu',
        'm_0': 'emu',
        't_offset': 's'
    }

    #: Model Parameter Unit mathmode strings
    UNITS_MM = {
        'tau*': r'$\mathregular{s}$',
        'beta': '',
        'm_eq': r'$\mathregular{emu}$',
        'm_0': r'$\mathregular{emu}$',
        't_offset': r'$\mathregular{s}$',
    }

    def __init__(self, fit_vars: dict[str, float | str],
                 fix_vars: dict[str, float | str], experiment: Experiment):

        # Replace any 'guess' strings with proper guesses
        self.fit_vars = self.set_initial_vals(fit_vars, experiment)
        self.fix_vars = self.set_initial_vals(fix_vars, experiment)

        # Check all PARNAMES are provided in fit+fix
        input_names = [
            name for name in {**self.fit_vars, **self.fix_vars}.keys()
        ]

        if any([req_name not in input_names for req_name in self.PARNAMES]):
            raise ValueError(
                'Missing fit/fix parameters in Exponential Model'
            )

        # Initialise remaining attributes required by Model superclass
        super().__init__()

        return

    @staticmethod
    def set_initial_vals(param_dict: dict[str, str | float],
                         experiment: Experiment):
        '''
        Sets guess values for parameters if requested by user

        Parameters
        ----------
        param_dict: dict[str, str | float]
            Keys are fit/fix parameter names (see class.PARNAMES)
            values are either float (actual value) or the string 'guess'\n
            If 'guess' then a parameter value is guessed using experiment
        experiment : Experiment
            Used to set guess values if specified

        Returns
        -------
        dict[str, float]
            Keys are fit/fix parameter names (see class.PARNAMES)
            values are float (actual value) which are initial values of
            parameter
        '''

        # Make copy, any str values will be replaced
        new_param_dict = copy.copy(param_dict)

        # Guesses
        guessdict = {
            'tau*': 100.,
            'm_eq': experiment.moments[-1],  # M_eq guess final measured moment
            'm_0': experiment.moments[0],  # M_0 is first measured moment
            'beta': 0.95,
            't_offset': 0.
        }

        # Replace 'guess' with relevant guess
        for var, val in param_dict.items():
            if val == 'guess':
                new_param_dict[var] = guessdict[var]

        return new_param_dict

    @staticmethod
    def model(parameters: dict[str, float], time: list[float]) -> list[float]:
        '''
        Evaluates exponential model function of DC magnetisation decay
        using provided parameter and time values.

        Parameters
        ----------
        parameters: dict[str, float]
            Parameters to use in model function, keys are given in
            ExponentialModel.PARNAMES
        time: list[float]
            time values in seconds at which model function is evaluated

        Returns
        -------
        list[float]
            moment as a function of time

        '''

        m_eq = parameters['m_eq']
        m_0 = parameters['m_0']
        beta = parameters['beta']
        tau_star = parameters['tau*']
        t_offset = parameters['t_offset']

        mom = m_eq
        mom += (m_0 - m_eq) * np.exp(-((time - t_offset) / tau_star)**beta)

        return mom

    def calc_lntau_expect(self) -> float:
        '''
        Calculates expectation value of ln(tau)

        Parameters
        ----------
        None

        Returns
        -------
        float
            <ln(tau)> value in ln(seconds)

        Raises
        -----
        ValueError
            If required model parameters in self.final_var_values are undefined
        '''

        beta = self.final_var_values['beta']
        tau_star = self.final_var_values['tau*']

        if None in [beta, tau_star]:
            _error = 'Cannot calculate ln(tau) expectation value '
            _error += 'beta and/or tau* are undefined!'
            raise ValueError(_error)

        return (1 - (1 / beta)) * np.euler_gamma + np.log(tau_star)

    def calc_lntau_stdev(self) -> float:
        '''
        Calculates standard deviation of ln(tau)

        Parameters
        ----------
        None

        Returns
        -------
        float
            Standard deviation of ln(tau) in ln(seconds)

        Raises
        -----
        ValueError
            If required model parameters in self.final_var_values are undefined
        '''

        beta = self.final_var_values['beta']

        if beta is None:
            _error = 'Error: Cannot calculate ln(tau) standard deviation'
            _error += 'beta undefined!'
            raise ValueError(_error)

        return np.sqrt((1. / beta**2 - 1.) * np.pi**2 / 6.)

    def calc_lntau_fit_ul(self) -> list[float]:
        '''
        Calculates upper and lower bounds of ln(tau) from fit uncertainty
        in fitted parameters

        Parameters
        ----------
        None

        Returns
        -------
        list[float]
            upper and lower bounds of ln(tau) from fit uncertainty in fitted
            parameters (upper > lower)
        '''

        tau_star = self.final_var_values['tau*']

        if 'tau*' in self.fit_stdev:
            tau_star_std = self.fit_stdev['tau*']
        else:
            tau_star_std = 0.

        beta = self.final_var_values['beta']

        if 'beta' in self.fit_stdev:
            beta_std = self.fit_stdev['beta']
        else:
            beta_std = 0.

        upper = (1 - (1 / (beta - beta_std))) * np.euler_gamma + np.log(tau_star + tau_star_std) # noqa
        lower = (1 - (1 / (beta + beta_std))) * np.euler_gamma + np.log(tau_star - tau_star_std) # noqa

        return sorted([upper, lower], reverse=True)


class DoubleExponentialModel(Model):
    '''
    Double Exponential Model of Magnetisation decay
    '''

    #: Model Name
    NAME = 'Double Exponential'

    #: Display name for interactive buttons
    DISP_NAME = copy.copy(NAME)

    #: Model Parameter name strings
    PARNAMES = [
        'tau*1', 'tau*2', 'beta1', 'beta2', 'm_eq', 'm_0', 't_offset', 'frac'
    ]

    #: Model Parameter bounds
    BOUNDS = {
        'tau*1': [0., np.inf],
        'tau*2': [0., np.inf],
        'beta1': [0., 1.],
        'beta2': [0., 1.],
        'frac': [0., 1.],
        'm_eq': [-np.inf, np.inf],
        'm_0': [0., np.inf],
        't_offset': [0., np.inf]
    }

    #: Model Parameter mathmode name strings
    VARNAMES_MM = {
        'tau*1': r'$\tau_{1}^*$',
        'tau*2': r'$\tau_{2}^*$',
        'lntau_expect1': r'$\langle \ln \tau 1 \rangle$',
        'lntau_expect2': r'$\langle \ln \tau 2 \rangle$',
        'lntau_stdev1': r'$\sigma_{\ln \tau 1}$',
        'lntau_stdev2': r'$\sigma_{\ln \tau 2}$',
        'beta1': r'$\beta_{1}$',
        'beta2': r'$\beta_{2}$',
        'frac': r'Fraction',
        'm_eq': r'$M_\mathregular{eq}$',
        'm_0': r'$M_0$',
        't_offset': r'$\mathregular{t}_\mathregular{offset}$'
    }

    #: Model Parameter Unit strings
    UNITS = {
        'tau*1': 's',
        'tau*2': 's',
        'beta1': '',
        'beta2': '',
        'frac': '',
        'm_eq': 'emu',
        'm_0': 'emu',
        't_offset': 's'
    }

    #: Model Parameter Unit mathmode strings
    UNITS_MM = {
        'tau*1': r'$\mathregular{s}$',
        'tau*2': r'$\mathregular{s}$',
        'beta1': '',
        'beta2': '',
        'frac': '',
        'm_eq': r'$\mathregular{emu}$',
        'm_0': r'$\mathregular{emu}$',
        't_offset': r'$\mathregular{s}$',
    }

    # Redefined for list
    @property
    def lntau_stdev(self) -> list[float]:
        'Standard deviation of ln(tau)'
        # If not calculated yet, then calculate
        if None in self._lntau_stdev:
            self._lntau_stdev = self.calc_lntau_stdev()
        return self._lntau_stdev

    # Redefined for list
    @property
    def lntau_fit_ul(self) -> list[list[float]]:
        '''
        Upper and lower (1 sigma) limits of ln(tau) from fit uncertainty
        in fitted values
        '''
        # If not calculated yet, then calculate
        if None in self._lntau_fit_ul:
            self._lntau_fit_ul = self.calc_lntau_fit_ul()
        return self._lntau_fit_ul

    # Redefined for list
    @property
    def lntau_expect(self) -> list[float]:
        'Expectation value of ln(tau)'
        # If not calculated yet, then calculate
        if None in self._lntau_expect:
            self._lntau_expect = self.calc_lntau_expect()
        return self._lntau_expect

    def __init__(self, fit_vars: dict[str, float | str],
                 fix_vars: dict[str, float | str], experiment: Experiment):

        # Replace any 'guess' strings with proper guesses
        self.fit_vars = self.set_initial_vals(fit_vars, experiment)
        self.fix_vars = self.set_initial_vals(fix_vars, experiment)

        # Check all PARNAMES are provided in fit+fix
        input_names = [
            name for name in {**self.fit_vars, **self.fix_vars}.keys()
        ]

        if any([req_name not in input_names for req_name in self.PARNAMES]):
            raise ValueError(
                'Missing fit/fix parameters in Double Exponential Model'
            )

        # Initialise remaining attributes required by Model superclass
        super().__init__()

        # Expectation value and standard deviation of ln(tau)
        self._lntau_expect = [None, None]
        self._lntau_fit_ul = [None, None]
        self._lntau_stdev = [None, None]

        return

    @staticmethod
    def set_initial_vals(param_dict: dict[str, str | float],
                         experiment: Experiment):
        '''
        Sets guess values for parameters if requested by user

        Parameters
        ----------
        param_dict: dict[str, str | float]
            Keys are fit/fix parameter names (see class.PARNAMES)
            values are either float (actual value) or the string 'guess'\n
            If 'guess' then a parameter value is guessed using experiment
        experiment : Experiment
            Used to set guess values if specified

        Returns
        -------
        dict[str, float]
            Keys are fit/fix parameter names (see class.PARNAMES)
            values are float (actual value) which are initial values of
            parameter
        '''

        # Make copy, any str values will be replaced
        new_param_dict = copy.copy(param_dict)

        # Guesses
        guessdict = {
            'tau*1': 50.,
            'tau*2': 5000.,
            'm_eq': experiment.moments[-1],  # M_eq guess final measured moment
            'm_0': experiment.moments[0],  # M_0 is first measured moment
            'frac': 0.5,
            'beta1': 0.95,
            'beta2': 0.95,
            't_offset': 0.
        }

        # Replace 'guess' with relevant guess
        for var, val in param_dict.items():
            if val == 'guess':
                new_param_dict[var] = guessdict[var]

        return new_param_dict

    @staticmethod
    def model(parameters: dict[str, float], time: list[float]) -> list[float]:
        '''
        Evaluates exponential model function of DC magnetisation decay
        using provided parameter and time values.

        Parameters
        ----------
        parameters: dict[str, float]
            Parameters to use in model function, keys are given in
            ExponentialModel.PARNAMES
        time: list[float]
            time values in seconds at which model function is evaluated

        Returns
        -------
        list[float]
            Moment as a function of time

        '''

        m_eq = parameters['m_eq']
        m_0 = parameters['m_0']
        beta1 = parameters['beta1']
        beta2 = parameters['beta2']
        tau_star1 = parameters['tau*1']
        tau_star2 = parameters['tau*2']
        frac = parameters['frac']
        t_offset = parameters['t_offset']

        sef1 = np.exp(-((time - t_offset) / tau_star1)**beta1)
        sef2 = np.exp(-((time - t_offset) / tau_star2)**beta2)
        mom = m_eq + (m_0 - m_eq) * ((frac * sef1) + ((1 - frac) * sef2))

        return mom

    def calc_lntau_expect(self) -> float:
        '''
        Calculates expectation value of ln(tau)

        Parameters
        ----------
        None

        Returns
        -------
        float
            <ln(tau)> value in ln(seconds)

        Raises
        -----
        ValueError
            If required model parameters in self.final_var_values are undefined
        '''

        beta1 = self.final_var_values['beta1']
        beta2 = self.final_var_values['beta2']
        tau_star1 = self.final_var_values['tau*1']
        tau_star2 = self.final_var_values['tau*2']

        if None in [beta1, beta2, tau_star1, tau_star2]:
            _error = 'Cannot calculate ln(tau) expectation value '
            _error += 'beta and/or tau* are undefined!'
            raise ValueError(_error)

        lntau1 = (1 - (1 / beta1)) * np.euler_gamma + np.log(tau_star1)
        lntau2 = (1 - (1 / beta2)) * np.euler_gamma + np.log(tau_star2)

        return [lntau1, lntau2]

    def calc_lntau_stdev(self) -> float:
        '''
        Calculates standard deviation of ln(tau1) and ln(tau2)

        Parameters
        ----------
        None

        Returns
        -------
        float
            Standard deviation of ln(tau) in ln(seconds)

        Raises
        -----
        ValueError
            If required model parameters in self.final_var_values are undefined
        '''

        beta1 = self.final_var_values['beta1']
        beta2 = self.final_var_values['beta2']

        if None in [beta1, beta2]:
            _error = 'Error: Cannot calculate ln(tau) standard deviation'
            _error += 'beta undefined!'
            raise ValueError(_error)

        sd1 = np.sqrt((1. / beta1**2 - 1.) * np.pi**2 / 6.)
        sd2 = np.sqrt((1. / beta2**2 - 1.) * np.pi**2 / 6.)

        return [sd1, sd2]

    def calc_lntau_fit_ul(self) -> list[float]:
        '''
        Calculates upper and lower bounds of ln(tau) from fit uncertainty
        in fitted parameters

        Parameters
        ----------
        None

        Returns
        -------
        list[float]
            upper and lower bounds of ln(tau) from fit uncertainty in fitted
            parameters (upper > lower)
        '''

        tau_star1 = self.final_var_values['tau*1']

        if 'tau*1' in self.fit_stdev:
            tau_star1_std = self.fit_stdev['tau*1']
        else:
            tau_star1_std = 0.

        beta1 = self.final_var_values['beta1']

        if 'beta1' in self.fit_stdev:
            beta1_std = self.fit_stdev['beta1']
        else:
            beta1_std = 0.

        tau_star2 = self.final_var_values['tau*2']

        if 'tau*2' in self.fit_stdev:
            tau_star2_std = self.fit_stdev['tau*2']
        else:
            tau_star2_std = 0.

        beta2 = self.final_var_values['beta2']

        if 'beta2' in self.fit_stdev:
            beta2_std = self.fit_stdev['beta2']
        else:
            beta2_std = 0.

        warnings.filterwarnings('ignore', 'invalid value encountered in log')

        upper1 = (1 - (1 / (beta1 - beta1_std))) * np.euler_gamma + np.log(tau_star1 + tau_star1_std) # noqa
        lower1 = (1 - (1 / (beta1 + beta1_std))) * np.euler_gamma + np.log(tau_star1 - tau_star1_std) # noqa

        upper2 = (1 - (1 / (beta2 - beta2_std))) * np.euler_gamma + np.log(tau_star2 + tau_star2_std) # noqa
        lower2 = (1 - (1 / (beta2 + beta2_std))) * np.euler_gamma + np.log(tau_star2 - tau_star2_std) # noqa

        warnings.filterwarnings('default', 'invalid value encountered in log')

        ul1 = sorted([upper1, lower1], reverse=True)
        ul2 = sorted([upper2, lower2], reverse=True)

        return [ul1, ul2]


def write_model_params(models: list[Model], file_name: str,
                       verbose: bool = True) -> None:
    '''
    Writes fitted parameters of a set of models to file.\n
    Assumes models are all of the same type, e.g. all Exponential

    Parameters
    ----------
    models : list[Model]
        Models, one per temperature, must all be same type
    file_name : str
        Name of output file
    verbose: bool, default True
        If True, file location is printed to terminal

    Returns
    -------
    None
    '''

    f = open(file_name, 'w', encoding='utf-8')

    f.write(' {:^12} '.format('T'))
    f.write('{:^12} '.format('H'))

    if all([model.dc_field != model.meas_dc_field for model in models]):
        f.write('{:^12} '.format('H_measured'))

    # Fitted parameters
    for name in models[0].fit_vars.keys():
        f.write('{:^17} {:^17} '.format(name, name + '-s-dev'))

    # Fixed parameters
    for name in models[0].fix_vars.keys():
        f.write('{:^17} '.format(name))

    if all({isinstance(model, DoubleExponentialModel) for model in models}):
        model_type = DoubleExponentialModel
    else:
        model_type = None

    if model_type == DoubleExponentialModel:
        f.write('{:^17} '.format('<ln(tau1)>'))
        f.write('{:^17} '.format('sigma_ln(tau1)'))
        f.write('{:^17} '.format('fit_upper_ln(tau1)'))
        f.write('{:^17} '.format('fit_lower_ln(tau1)'))
        f.write('{:^17} '.format('<ln(tau2)>'))
        f.write('{:^17} '.format('sigma_ln(tau2)'))
        f.write('{:^17} '.format('fit_upper_ln(tau2)'))
        f.write('{:^17} '.format('fit_lower_ln(tau2)'))
    else:
        f.write('{:^17} '.format('<ln(tau)>'))
        f.write('{:^17} '.format('sigma_ln(tau)'))
        f.write('{:^17} '.format('fit_upper_ln(tau)'))
        f.write('{:^17} '.format('fit_lower_ln(tau)'))

    f.write('\n')

    for model in models:
        if not model.fit_status:
            continue

        f.write('{:12.10f} '.format(model.temperature))
        f.write('{:12.10f} '.format(model.dc_field))

        if model.dc_field != model.meas_dc_field:
            f.write('{:12.10f} '.format(model.meas_dc_field))

        for name in model.fit_vars.keys():
            f.write('{: 1.10E} {: 1.10E} '.format(
                model.final_var_values[name], model.fit_stdev[name]
            ))

        for value in model.fix_vars.values():
            f.write('{: 1.10E} '.format(value))

        if model_type == DoubleExponentialModel:
            f.write('{: 1.10E} '.format(model.lntau_expect[0]))
            f.write('{: 1.10E} '.format(model.lntau_stdev[0]))
            f.write('{: 1.10E} '.format(model.lntau_fit_ul[0][0]))
            f.write('{: 1.10E} '.format(model.lntau_fit_ul[0][1]))
            f.write('{: 1.10E} '.format(model.lntau_expect[1]))
            f.write('{: 1.10E} '.format(model.lntau_stdev[1]))
            f.write('{: 1.10E} '.format(model.lntau_fit_ul[1][0]))
            f.write('{: 1.10E} '.format(model.lntau_fit_ul[1][1]))
        else:
            f.write('{: 1.10E} '.format(model.lntau_expect))
            f.write('{: 1.10E} '.format(model.lntau_stdev))
            f.write('{: 1.10E} '.format(model.lntau_fit_ul[0]))
            f.write('{: 1.10E} '.format(model.lntau_fit_ul[1]))

        f.write('\n')

    if verbose:
        ut.cprint(
            f'\n Model parameters written to \n {file_name}\n',
            'cyan'
        )
    return


def write_model_data(experiments: list['Experiment'],
                     models: list['Model'], file_name: str,
                     verbose: bool = True) -> None:
    '''
    Creates file containing time, and moment using the model
    function with fitted and fixed parameters.
    Temperatures for which a fit was not possible are not included.

    Parameters
    ----------
    experiments : list[Experiment]
        List of experiments to which a model was fitted
    models : list[Model]
        List of models, one per experiment
    file_name: str
        Name of output file
    verbose: bool, default True
        If True, file location is printed to terminal

    Returns
    -------
    None
    '''

    f = open(file_name, 'w', encoding='utf-8')

    # For each experiment and its corresponding model, calculate
    # moment at experimental time values
    for model, experiment in zip(models, experiments):

        moments = model.model(
            model.final_var_values,
            experiment.times,
        )

        f.write('{:11} {:.5f} K\n'.format('T = ', model.temperature))
        f.write('{:11} {:.5f} Oe\n'.format('H = ', model.dc_field))

        f.write('{:^15} {:^15}\n'.format(
            'Time (s)', 'Moment (emu)'
        ))
        for t, mo in zip(experiment.times, moments):
            f.write('{: 1.8E} {: 1.8E}\n'.format(t, mo))

    if verbose:
        ut.cprint(
            f'\n Model decays written to \n {file_name}\n',
            'cyan'
        )
    return


def plot_decays_and_fields(experiments: list[Experiment], show: bool = True,
                           save: bool = False,
                           save_name: str = 'decays_and_fields.png',
                           x_scale: str = 'linear', y_scale: str = 'linear',
                           verbose: bool = True) -> tuple[plt.Figure, list[plt.Axes]]: # noqa:
    '''
    Plots a list of experimental decays as moment vs time and field vs time,
    on one figure with two plots.

    Parameters
    ----------
    experiment: list[Experiment]
        Experiments to plot
    show: bool, default True
        If True, shows plot on screen
    save: bool, default False
        If True, saves plot to file
    save_name : str, default 'decays_and_fields.png'
        If save is True, saves plot to this file name
    x_scale : str {'linear', 'log'}
        Scale of x-axis
    y_scale : str {'linear', 'log'}
        Scale of y-axis
    verbose: bool, default True
        If True, plot file location is written to terminal

    Returns
    -------
    plt.Figure
        Matplotlib figure object
    list[plt.Axes]
        List of Matplotlib axis objects
    '''

    fig, (ax1, ax2) = plt.subplots(1, 2, sharex=True, figsize=(10, 5))
    for experiment in experiments:

        experiment.times -= np.min(experiment.times)
        ax1.plot(
            experiment.times,
            experiment.moments,
            lw=2,
            marker='x',
            label='{:.2f} K'.format(experiment.rep_temperature)
        )
        ax2.plot(
            experiment.times,
            experiment.dc_fields,
            lw=2,
            marker='x',
            label='{:.2f} K'.format(experiment.rep_temperature)
        )

    ax1.set_xlabel('Time (s)')
    ax1.set_ylabel('Moment (emu)')
    ax2.set_xlabel('Time (s)')
    ax2.set_ylabel('DC Field (Oe)')

    if len(np.unique([e.rep_temperature for e in experiments])) > 1:
        ax1.legend()
        ax2.legend()
    else:
        fig.suptitle(
            'T = {:.2f} K'.format(
                experiments[0].rep_temperature
            ),
            fontsize='medium'
        )

    ax1.set_xscale(x_scale)
    ax1.set_yscale(y_scale)

    if y_scale == 'linear':
        ax1.yaxis.set_minor_locator(AutoMinorLocator())
    if x_scale == 'linear':
        ax1.xaxis.set_minor_locator(AutoMinorLocator())

    ax2.xaxis.set_minor_locator(AutoMinorLocator())
    ax2.yaxis.set_minor_locator(AutoMinorLocator())

    fig.tight_layout()

    if save:
        plt.savefig(save_name, dpi=500)
        if verbose:
            ut.cprint(
                f'\n Decay plot saved to \n {save_name}\n',
                'cyan'
            )
    if show:
        plt.show()

    plt.close('all')

    return fig, [ax1, ax2]


def plot_decays(experiments: list[Experiment], show: bool = True,
                save: bool = False, save_name: str = 'decays.png',
                x_scale: str = 'linear', y_scale: str = 'linear',
                verbose: bool = True) -> tuple[plt.Figure, plt.Axes]:
    '''
    Plots a list of experimental decays (moment vs time) on one plot.

    Parameters
    ----------
    experiment: list[Experiment]
        Experiments to plot
    show: bool, default True
        If True, shows plot on screen
    save: bool, default False
        If True, saves plot to file
    save_name : str, default 'decays.png'
        If save is True, saves plot to this file name
    x_scale : str {'linear', 'log'}
        Scale of x-axis
    y_scale : str {'linear', 'log'}
        Scale of y-axis
    verbose: bool, default True
        If True, plot file location is written to terminal

    Returns
    -------
    plt.Figure
        Matplotlib figure object
    plt.Axes
        Matplotlib axis object
    '''

    fig, ax = plt.subplots(1, 1, figsize=(6, 5))
    for experiment in experiments:

        experiment.times -= np.min(experiment.times)
        ax.plot(
            experiment.times,
            experiment.moments,
            lw=2,
            marker='x',
            label='{:.2f} K'.format(experiment.rep_temperature)
        )

    ax.set_xlabel('Time (s)')
    ax.set_ylabel('Moment (emu)')

    if len(np.unique([e.rep_temperature for e in experiments])) > 1:
        ax.legend()
    else:
        fig.suptitle(
            'T = {:.2f} K'.format(
                experiments[0].rep_temperature
            ),
            fontsize='medium'
        )

    if y_scale == 'linear':
        ax.yaxis.set_minor_locator(AutoMinorLocator())
    if x_scale == 'linear':
        ax.xaxis.set_minor_locator(AutoMinorLocator())

    ax.set_xscale(x_scale)

    fig.tight_layout()

    if save:
        plt.savefig(save_name, dpi=500)
        if verbose:
            ut.cprint(
                f'\n Decay plot saved to \n {save_name}\n',
                'cyan'
            )
    if show:
        plt.show()

    plt.close('all')

    return fig, ax


def plot_decay(experiment: Experiment, show: bool = True,
               save: bool = False, save_name: str = 'decay.png',
               x_scale: str = 'linear', y_scale: str = 'linear',
               verbose: bool = True) -> tuple[plt.Figure, plt.Axes]:
    '''
    Plots an experimental decay (moment vs time) on one plot.

    Parameters
    ----------
    experiment: Experiment
        Experiment to plot
    show: bool, default True
        If True, shows plot on screen
    save: bool, default False
        If True, saves plot to file
    save_name : str, default 'decay.png'
        If save is True, saves plot to this file name
    x_scale : str {'linear', 'log'}
        Scale of x-axis
    y_scale : str {'linear', 'log'}
        Scale of y-axis
    verbose: bool, default True
        If True, plot file location is written to terminal

    Returns
    -------
    plt.Figure
        Matplotlib figure object
    plt.Axes
        Matplotlib axis object
    '''

    fig, ax = plt.subplots(1, 1, figsize=(6, 5))
    experiment.times -= np.min(experiment.times)
    ax.plot(
        experiment.times,
        experiment.moments,
        lw=2,
        marker='x',
        label='{:.2f} K'.format(experiment.rep_temperature)
    )

    ax.set_xlabel('Time (s)')
    ax.set_ylabel('Moment (emu)')

    if y_scale == 'linear':
        ax.yaxis.set_minor_locator(AutoMinorLocator())
    if x_scale == 'linear':
        ax.xaxis.set_minor_locator(AutoMinorLocator())

    ax.set_xscale(x_scale)

    fig.tight_layout()

    if save:
        plt.savefig(save_name, dpi=500)
        if verbose:
            ut.cprint(
                f'\n Decay plot saved to \n {save_name}\n',
                'cyan'
            )
    if show:
        plt.show()

    plt.close('all')

    return fig, ax


def plot_fitted_decay(experiment: Experiment, model: Model,
                      show: bool = True, save: bool = False,
                      save_name: str = 'fitted_decay.png',
                      show_params: bool = True,
                      x_scale: str = 'linear', y_scale: str = 'linear',
                      verbose: bool = True) -> tuple[plt.Figure, plt.Axes]:
    '''
    Plots experimental and fitted (model) decay together (moment vs time),
    displays on screen.

    Parameters
    ----------
    experiment: Experiment
        Experiment to plot
    model: Model
        Model (fitted) to plot
    show: bool, default True
        If True, shows plot on screen
    save: bool, default False
        If True, saves plot to file
    save_name : str, default 'fitted_decay.png'
        If save is True, saves plot to this file name
    show_params: bool, default True
        If True, shows fitted parameters on decay plots
    x_scale : str {'linear', 'log'}
        Scale of x-axis
    y_scale : str {'linear', 'log'}
        Scale of y-axis
    verbose: bool, default True
        If True, plot file location is written to terminal

    Returns
    -------
    plt.Figure
        Matplotlib figure object
    plt.Axes
        Matplotlib axis object
    '''

    if show_params:
        figsize = (4.5, 5)
    else:
        figsize = (4.5, 4)

    # Plot the figure.
    fig, ax = plt.subplots(
        1,
        1,
        sharex=True,
        sharey=True,
        figsize=figsize,
        num='Fitted DC Decay at {:6.2f} K and {:6.2f} Oe'.format(
            experiment.rep_temperature, experiment.dc_fields[-1]
        )
    )

    expression = ''

    for name in model.PARNAMES:
        if np.abs(model.final_var_values[name]) > 1E-8:
            # Meq format as x.yz x 10^a
            if name == 'm_eq':
                # Convert to scientific string, then split at E
                # to get mantissa and exponent separately
                _tmp = f'{model.final_var_values[name]:.4E}'
                _tmp = _tmp.split('E')
                _mant = float(_tmp[0])
                _exp = int(_tmp[1])
                expression += f'{model.VARNAMES_MM[name]} = {_mant:.4f}'
                expression += r' $\times \ \mathregular{10}'
                expression += r'^\mathregular{{{:d}}}$'.format(_exp)
                if 'm_eq' in model.fit_vars.keys():
                    expression += r' $\pm$'
                    _tmp = f'{model.fit_stdev[name]:.4E}'
                    _tmp = _tmp.split('E')
                    _mant = float(_tmp[0])
                    _exp = int(_tmp[1])
                    expression += f' {_mant:.4f}'
                    expression += r' $\times \ \mathregular{10}'
                    expression += r'^\mathregular{{{:d}}}$'.format(_exp)
            else:
                # All others format as float
                expression += '{} = {:.4f} '.format(
                    model.VARNAMES_MM[name],
                    model.final_var_values[name],
                )
                if name in model.fit_vars.keys():
                    expression += r'$\pm$'
                    expression += ' {:.4f}'.format(model.fit_stdev[name])
            expression += f' {model.UNITS_MM[name]} '
            expression += '\n'

    if show_params:
        ax.text(0.0, 1.02, s=expression, fontsize=10, transform=ax.transAxes)

    label_fit = 'Fit'

    ax.plot(
        experiment.times,
        experiment.moments,
        lw=0,
        marker='o',
        fillstyle='none',
        label='Experiment'
    )
    ax.plot(
        experiment.times,
        model.model(
            model.final_var_values,
            experiment.times,
        ),
        lw=1.5,
        label=label_fit
    )

    ax.legend(loc=0, fontsize='10', numpoints=1, ncol=1, frameon=False)
    ax.set_ylabel('Moment (emu)')
    ax.set_xlabel('Time (s)')

    ax.set_xscale(x_scale)
    ax.set_yscale(y_scale)

    if y_scale == 'linear':
        ax.yaxis.set_minor_locator(AutoMinorLocator())
    if x_scale == 'linear':
        ax.xaxis.set_minor_locator(AutoMinorLocator())

    fig.tight_layout()

    if save:
        plt.savefig(save_name, dpi=500)
        if verbose:
            ut.cprint(
                f'\n Fitted decay plot saved to \n {save_name}\n',
                'cyan'
            )
    if show:
        plt.show()

    plt.close()
    return fig, ax
